from __future__ import annotations

import logging
from typing import Tuple, Dict, Iterator, TYPE_CHECKING

import numpy
from amulet_nbt import (
    ByteTag,
    IntTag,
    LongTag,
    StringTag,
    ListTag,
    CompoundTag,
    ByteArrayTag,
    IntArrayTag,
    AbstractBaseArrayTag,
    NamedTag,
)

import amulet
from amulet.api.data_types import SubChunkNDArray, AnyNDArray, BlockCoordinates
from amulet.utils import world_utils
from amulet.api.wrapper import EntityIDType, EntityCoordType
from amulet.api.chunk import Chunk, StatusFormats
from .base_anvil_interface import (
    BaseAnvilInterface,
    ChunkDataType,
    ChunkPathType,
)

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk

log = logging.getLogger(__name__)


class AnvilNAInterface(BaseAnvilInterface):
    Level: ChunkPathType = ("region", [("Level", CompoundTag)], CompoundTag)
    Sections: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Sections", ListTag)],
        ListTag,
    )

    BlockEntities: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("TileEntities", ListTag)],
        ListTag,
    )
    Entities: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Entities", ListTag)],
        ListTag,
    )
    InhabitedTime: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("InhabitedTime", LongTag)],
        LongTag,
    )
    LastUpdate: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("LastUpdate", LongTag)],
        LongTag,
    )
    HeightMap: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("HeightMap", IntArrayTag)],
        IntArrayTag,
    )
    TerrainPopulated: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("TerrainPopulated", ByteTag)],
        ByteTag,
    )
    LightPopulated: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("LightPopulated", ByteTag)],
        ByteTag,
    )
    V: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("V", ByteTag)],
        ByteTag(1),
    )
    BlockTicks: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("TileTicks", ListTag)],
        ListTag,
    )
    Biomes: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Biomes", ByteArrayTag)],
        None,
    )
    xPos: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("xPos", IntTag)],
        IntTag,
    )
    zPos: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("zPos", IntTag)],
        IntTag,
    )

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "256IA")

        self._set_feature("light_optional", "false")

        self._set_feature("block_entity_format", EntityIDType.namespace_str_id)
        self._set_feature("block_entity_coord_format", EntityCoordType.xyz_int)

        self._set_feature("entity_format", EntityIDType.namespace_str_id)
        self._set_feature("entity_coord_format", EntityCoordType.Pos_list_double)

        self._register_decoder(self._decode_coords)
        self._register_decoder(self._decode_last_update)
        self._register_decoder(self._decode_status)
        self._register_decoder(self._decode_v_tag)
        self._register_decoder(self._decode_inhabited_time)
        self._register_decoder(self._decode_biomes)
        self._register_decoder(self._decode_height)
        self._register_decoder(self._decode_entities)
        self._register_decoder(self._decode_blocks)
        self._register_decoder(self._decode_block_entities)
        self._register_decoder(self._decode_block_ticks)
        self._register_decoder(self._decode_block_light)
        self._register_decoder(self._decode_sky_light)

        self._register_encoder(self._encode_coords)
        self._register_encoder(self._encode_last_update)
        self._register_encoder(self._encode_status)
        self._register_encoder(self._encode_v_tag)
        self._register_encoder(self._encode_inhabited_time)
        self._register_encoder(self._encode_biomes)
        self._register_encoder(self._encode_height)
        self._register_encoder(self._encode_entities)
        self._register_encoder(self._encode_blocks)
        self._register_encoder(self._encode_block_entities)
        self._register_encoder(self._encode_block_ticks)
        self._register_encoder(self._encode_block_light)
        self._register_encoder(self._encode_sky_light)

        self._register_post_encoder(self._post_encode_sections)

    @staticmethod
    def minor_is_valid(key: int):
        return key == -1

    def decode(
        self, cx: int, cz: int, data: ChunkDataType, bounds: Tuple[int, int]
    ) -> Tuple["Chunk", AnyNDArray]:
        chunk = self._init_decode(cx, cz, data)
        floor_cy = self._get_floor_cy(data)
        height_cy = (bounds[1] - bounds[0]) >> 4
        self._do_decode(chunk, data, floor_cy, height_cy)
        block_palette = chunk.misc.pop("block_palette")
        return chunk, block_palette

    def _get_floor_cy(self, data: ChunkDataType):
        return 0

    @staticmethod
    def _init_decode(cx: int, cz: int, data: ChunkDataType) -> Chunk:
        """Get the decode started by creating a chunk object."""
        chunk = Chunk(cx, cz)
        chunk.misc = {
            # store the chunk data so that any non-versioned data can get saved back
            "_java_chunk_data_layers": data
        }
        return chunk

    def _get_level(self, data: ChunkDataType) -> CompoundTag:
        """
        Get the level data container
        For older levels this is region:Level but newer worlds it is in region root
        :param data: The raw chunk data
        :return: The level data compound
        """
        return self.get_layer_obj(data, self.Level)

    def _decode_coords(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.get_layer_obj(data, self.xPos, pop_last=True)
        self.get_layer_obj(data, self.zPos, pop_last=True)

    def _decode_last_update(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["last_update"] = self.get_layer_obj(
            data, self.LastUpdate, pop_last=True
        ).py_int

    def _decode_status(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        status = "empty"
        if self.get_layer_obj(data, self.TerrainPopulated, pop_last=True):
            status = "decorated"
        if self.get_layer_obj(data, self.LightPopulated, pop_last=True):
            status = "postprocessed"
        chunk.status = status

    def _decode_v_tag(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["V"] = self.get_layer_obj(data, self.V, pop_last=True).py_int

    def _decode_inhabited_time(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["inhabited_time"] = self.get_layer_obj(
            data, self.InhabitedTime, pop_last=True
        ).py_int

    def _decode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        biomes = self.get_layer_obj(data, self.Biomes, pop_last=True)
        if isinstance(biomes, AbstractBaseArrayTag) and biomes.np_array.size == 256:
            chunk.biomes = biomes.np_array.astype(numpy.uint32).reshape((16, 16))

    def _decode_height(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        height = self.get_layer_obj(data, self.HeightMap, pop_last=True).np_array
        if isinstance(height, numpy.ndarray) and height.size == 256:
            chunk.misc["height_map256IA"] = height.reshape((16, 16))

    def _iter_sections(self, data: ChunkDataType) -> Iterator[Tuple[int, CompoundTag]]:
        sections: ListTag = self.get_layer_obj(data, self.Sections)
        for section in sections:
            yield section["Y"].py_int, section

    def _decode_blocks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        blocks: Dict[int, SubChunkNDArray] = {}
        palette = []
        palette_len = 0
        for cy, section in self._iter_sections(data):
            block_tag = section.pop("Blocks", None)
            data_tag = section.pop("Data", None)
            if not isinstance(block_tag, AbstractBaseArrayTag) or not isinstance(
                data_tag, AbstractBaseArrayTag
            ):
                continue
            section_blocks = numpy.asarray(block_tag, dtype=numpy.uint8)
            section_data = numpy.asarray(data_tag, dtype=numpy.uint8)
            section_blocks = section_blocks.reshape((16, 16, 16))
            section_blocks = section_blocks.astype(numpy.uint16)

            section_data = world_utils.from_nibble_array(section_data)
            section_data = section_data.reshape((16, 16, 16))

            add_tag = section.pop("Add", None)
            if isinstance(add_tag, AbstractBaseArrayTag):
                add_blocks = numpy.asarray(add_tag, dtype=numpy.uint8)
                add_blocks = world_utils.from_nibble_array(add_blocks)
                add_blocks = add_blocks.reshape((16, 16, 16))

                section_blocks |= add_blocks.astype(numpy.uint16) << 8
                # TODO: fix this

            (section_palette, blocks[cy]) = world_utils.fast_unique(
                numpy.transpose(
                    (section_blocks << 4) + section_data, (2, 0, 1)
                )  # YZX -> XYZ
            )
            blocks[cy] += palette_len
            palette_len += len(section_palette)
            palette.append(section_palette)

        if palette:
            final_palette, lut = numpy.unique(
                numpy.concatenate(palette), return_inverse=True
            )
            final_palette: numpy.ndarray = numpy.array(
                [final_palette >> 4, final_palette & 15]
            ).T
            for cy in blocks:
                blocks[cy] = lut[blocks[cy]]
        else:
            final_palette = numpy.array([], dtype=object)
        chunk.blocks = blocks
        chunk.misc["block_palette"] = final_palette

    def _unpack_light(
        self, data: ChunkDataType, section_key: str
    ) -> Dict[int, numpy.ndarray]:
        light_container = {}
        for cy, section in self._iter_sections(data):
            if self.check_type(section, section_key, ByteArrayTag):
                light: numpy.ndarray = section.pop(section_key).np_array
                if light.size == 2048:
                    # TODO: check if this needs transposing or if the values are the other way around
                    light_container[cy] = (
                        (
                            light.reshape(-1, 1)
                            & numpy.array([0xF, 0xF0], dtype=numpy.uint8)
                        )
                        >> numpy.array([0, 4], dtype=numpy.uint8)
                    ).reshape((16, 16, 16))
        return light_container

    def _decode_block_light(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["block_light"] = self._unpack_light(data, "BlockLight")

    def _decode_sky_light(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["sky_light"] = self._unpack_light(data, "SkyLight")

    def _decode_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        ents = self._decode_entity_list(
            self.get_layer_obj(data, self.Entities, pop_last=True)
        )
        if amulet.entity_support:
            chunk.entities = ents
        else:
            chunk._native_entities.extend(ents)
            chunk._native_version = ("java", -1)

    def _decode_block_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.block_entities = self._decode_block_entity_list(
            self.get_layer_obj(data, self.BlockEntities, pop_last=True)
        )

    @staticmethod
    def _decode_ticks(ticks: ListTag) -> Dict[BlockCoordinates, Tuple[str, int, int]]:
        return {
            (tick["x"].py_data, tick["y"].py_data, tick["z"].py_data): (
                tick["i"].py_data,
                tick["t"].py_data,
                tick["p"].py_data,
            )
            for tick in ticks
            if all(c in tick and isinstance(tick[c], IntTag) for c in "xyztp")
            and "i" in tick
            and isinstance(tick["i"], StringTag)
        }

    def _decode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc.setdefault("block_ticks", {}).update(
            self._decode_ticks(self.get_layer_obj(data, self.BlockTicks, pop_last=True))
        )

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, int],
        bounds: Tuple[int, int],
    ) -> ChunkDataType:
        floor_cy = bounds[0] >> 4
        height_cy = (bounds[1] - bounds[0]) >> 4
        data = self._init_encode(chunk, max_world_version, floor_cy, height_cy)
        chunk.misc["block_palette"] = palette
        self._do_encode(chunk, data, floor_cy, height_cy)
        return data

    def _init_encode(
        self,
        chunk: "Chunk",
        max_world_version: Tuple[str, int],
        floor_cy: int,
        height_cy: int,
    ) -> ChunkDataType:
        """Get or create the root data."""
        data = chunk.misc.get("_java_chunk_data_layers", None)
        if not isinstance(data, dict):
            data = {}
        return {
            key: value
            for key, value in data.items()
            if isinstance(key, str) and isinstance(value, NamedTag)
        }

    def _get_encode_sections(
        self,
        data: ChunkDataType,
        floor_cy: int,
        height_cy: int,
    ) -> Dict[int, CompoundTag]:
        """Get or create the section array populating all valid sections"""
        sections: ListTag = self.set_layer_obj(data, self.Sections, setdefault=True)
        section_map: Dict[int, CompoundTag] = {}
        section: CompoundTag
        for section_index in range(len(sections) - 1, -1, -1):
            section = sections[section_index]
            cy = section.get("Y", None)
            if isinstance(cy, ByteTag):
                section_map[cy.py_int] = section
            else:
                sections.pop(section_index)
        for cy in range(floor_cy, floor_cy + height_cy):
            if cy not in section_map:
                section = section_map[cy] = CompoundTag({"Y": ByteTag(cy)})
                sections.append(section)
        return section_map

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, CompoundTag],
        palette: AnyNDArray,
        cy: int,
    ):
        block_sub_array = palette[
            numpy.transpose(
                chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
            ).ravel()  # XYZ -> YZX
        ]

        data_sub_array = block_sub_array[:, 1]
        block_sub_array = block_sub_array[:, 0]
        # if not numpy.any(block_sub_array) and not numpy.any(data_sub_array):
        #     return False
        sections[cy]["Blocks"] = ByteArrayTag(block_sub_array.astype("uint8"))
        sections[cy]["Data"] = ByteArrayTag(world_utils.to_nibble_array(data_sub_array))

    def _encode_blocks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        sections = self._get_encode_sections(data, floor_cy, height_cy)
        block_palette = chunk.misc.pop("block_palette")
        ceil_cy = floor_cy + height_cy
        for cy in chunk.blocks.sub_chunks:
            if floor_cy <= cy < ceil_cy:
                self._encode_block_section(chunk, sections, block_palette, cy)

    def _encode_coords(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(data, self.xPos, IntTag(chunk.cx))
        self.set_layer_obj(data, self.zPos, IntTag(chunk.cz))

    def _encode_last_update(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data, self.LastUpdate, LongTag(chunk.misc.get("last_update", 0))
        )

    def _encode_status(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        status = chunk.status.as_type(StatusFormats.Raw)
        self.set_layer_obj(data, self.TerrainPopulated, ByteTag(int(status > -0.3)))
        self.set_layer_obj(data, self.LightPopulated, ByteTag(int(status > -0.2)))

    def _encode_v_tag(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(data, self.V, ByteTag(chunk.misc.get("V", 1)))

    def _encode_inhabited_time(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data, self.InhabitedTime, LongTag(chunk.misc.get("inhabited_time", 0))
        )

    def _encode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.biomes.convert_to_2d()
        self.set_layer_obj(
            data,
            self.Biomes,
            ByteArrayTag(chunk.biomes.astype(dtype=numpy.uint8).ravel()),
        )

    def _encode_height(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        height = chunk.misc.get("height_map256IA", None)
        if (
            isinstance(height, numpy.ndarray)
            and numpy.issubdtype(height.dtype, numpy.integer)
            and height.shape == (16, 16)
        ):
            self.set_layer_obj(
                data,
                self.HeightMap,
                IntArrayTag(numpy.zeros(256, dtype=numpy.uint32)),
            )
        elif self._features["height_map"] == "256IARequired":
            self.set_layer_obj(data, self.HeightMap, IntArrayTag(height.ravel()))

    def _encode_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        if amulet.entity_support:
            entities = chunk.entities
        else:
            entities = chunk._native_entities

        self.set_layer_obj(
            data,
            self.Entities,
            self._encode_entity_list(entities),
        )

    def _encode_block_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data,
            self.BlockEntities,
            self._encode_block_entity_list(chunk.block_entities),
        )

    @staticmethod
    def _encode_ticks(ticks: Dict[BlockCoordinates, Tuple[str, int, int]]) -> ListTag:
        ticks_out = ListTag()
        if isinstance(ticks, dict):
            for k, v in ticks.items():
                try:
                    (x, y, z), (i, t, p) = k, v
                    ticks_out.append(
                        CompoundTag(
                            {
                                "i": StringTag(i),
                                "p": IntTag(p),
                                "t": IntTag(t),
                                "x": IntTag(x),
                                "y": IntTag(y),
                                "z": IntTag(z),
                            }
                        )
                    )
                except Exception:
                    log.error(f"Could not serialise tick data {k}: {v}")
        return ticks_out

    def _encode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data, self.BlockTicks, self._encode_ticks(chunk.misc.get("block_ticks", {}))
        )

    def _pack_light(
        self,
        chunk: Chunk,
        data: ChunkDataType,
        floor_cy: int,
        height_cy: int,
        feature_key: str,
        section_key: str,
    ):
        light_container = chunk.misc.get(feature_key, {})
        if not isinstance(light_container, dict):
            light_container = {}
        for cy, section in self._get_encode_sections(data, floor_cy, height_cy).items():
            light = light_container.get(cy, None)
            if (
                isinstance(light, numpy.ndarray)
                and numpy.issubdtype(light.dtype, numpy.integer)
                and light.shape == (16, 16, 16)
            ):
                light = light.ravel() % 16
                section[section_key] = ByteArrayTag(light[::2] + (light[1::2] << 4))
            elif self._features["light_optional"] == "false":
                section[section_key] = ByteArrayTag(
                    numpy.full(2048, 255, dtype=numpy.uint8)
                )

    def _encode_block_light(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self._pack_light(chunk, data, floor_cy, height_cy, "block_light", "BlockLight")

    def _encode_sky_light(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self._pack_light(chunk, data, floor_cy, height_cy, "sky_light", "SkyLight")

    def _post_encode_sections(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        """Strip out all empty sections"""
        sections = self.get_layer_obj(data, self.Sections)
        if sections:
            for i in range(len(sections) - 1, -1, -1):
                section = sections[i]
                if "Blocks" not in section or "Data" not in section:
                    # in 1.12 if a section exists, Blocks and Data must exist
                    sections.pop(i)
        if not sections:
            # if no sections remain we can remove the sections data
            self.get_layer_obj(data, self.Sections, pop_last=True)


export = AnvilNAInterface



from __future__ import annotations

from sys import getsizeof
import re
from typing import Dict, Iterable, Tuple, Union, Mapping
from amulet_nbt import ByteTag, ShortTag, IntTag, LongTag, StringTag, from_snbt

from .errors import BlockException

PropertyValueType = Union[
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
    StringTag,
]
PropertyType = Mapping[str, PropertyValueType]
PropertyTypeMultiple = Dict[str, Tuple[PropertyValueType, ...]]

PropertyDataTypes = (
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
    StringTag,
)


class Block:
    """
    A class to manage the state of a block.

    It is an immutable object that contains a namespaced name, properties and extra blocks.

    Here's a few examples on how create a Block object:

    >>> # Create a block with the namespace `minecraft` and base name `stone`
    >>> stone = Block("minecraft", "stone")

    >>> # Create a water block with the level property
    >>> water = Block(
    >>>     "minecraft",  # the namespace
    >>>     "water",  # the base name
    >>>     {  # A dictionary of properties.
    >>>         # Keys must be strings and values must be a numerical or string NBT type.
    >>>         "level": StringTag("0")  # define a property `level` with a string value `1`
    >>>     }
    >>> )

    >>> # The above two examples can also be achieved by creating the block from the Java blockstate.
    >>> stone = Block.from_string_blockstate("minecraft:stone")
    >>> water = Block.from_string_blockstate("minecraft:water[level=0]")

    Java 1.13 added the concept of waterlogging blocks whereby some blocks have a `waterlogged` property.
    Bedrock achieved the same behaviour by added a layering system. The layering system allows any block to be the second block.
    In order to support both formats and future proof the editor Amulet adds the concept of extra blocks.
    Extra blocks are a sequence of zero or more blocks stored in the `extra_blocks` property of the Block instance.
    Amulet places no restrictions on which blocks can be extra blocks but some validation will be done when saving based on the format being saved to.

    >>> # Create a waterlogged stone block.
    >>> waterlogged_stone = Block("minecraft", "stone",
    >>>     # extra_blocks can be a Block instance or iterable of Block instances.
    >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
    >>> )

    >>> # The above can also be achieved by adding together a stone and water block.
    >>> waterlogged_stone = stone + water
    >>> repr(waterlogged_stone)
    'Block(minecraft:stone, extra_blocks=(minecraft:water[level="0"]))'
    """

    __slots__ = (
        "_namespaced_name",
        "_namespace",
        "_base_name",
        "_properties",
        "_extra_blocks",
        "_blockstate",
        "_snbt_blockstate",
        "_full_blockstate",
    )  # Reduces memory footprint

    snbt_blockstate_regex = re.compile(
        r"(?:(?P<namespace>[a-z0-9_.-]+):)?(?P<base_name>[a-z0-9/._-]+)(?:\[(?P<property_name>[a-z0-9_]+)=(?P<property_value>[a-z0-9_\"']+)(?P<properties>.*)\])?"
    )
    blockstate_regex = re.compile(
        r"(?:(?P<namespace>[a-z0-9_.-]+):)?(?P<base_name>[a-z0-9/._-]+)(?:\[(?P<property_name>[a-z0-9_]+)=(?P<property_value>[a-z0-9_]+)(?P<properties>.*)\])?"
    )

    snbt_properties_regex = re.compile(
        r"(?:,(?P<name>[a-z0-9_]+)=(?P<value>[a-z0-9_\"']+))"
    )
    properties_regex = re.compile(r"(?:,(?P<name>[a-z0-9_]+)=(?P<value>[a-z0-9_]+))")

    _extra_blocks: Tuple[Block, ...]

    def __init__(
        self,
        namespace: str,
        base_name: str,
        properties: PropertyType = None,
        extra_blocks: Union[Block, Iterable[Block]] = None,
    ):
        """
        Constructs a :class:`Block` instance.

        >>> stone = Block("minecraft", "stone")
        >>>
        >>> # Create a water block with the level property
        >>> water = Block(
        >>>     "minecraft",  # the namespace
        >>>     "water",  # the base name
        >>>     {  # A dictionary of properties.
        >>>         # Keys must be strings and values must be a numerical or string NBT type.
        >>>         "level": StringTag("0")  # define a property `level` with a string value `1`
        >>>     }
        >>> )

        :param namespace: The string namespace of the block. eg `minecraft`
        :param base_name: The string base name of the block. eg `stone`
        :param properties: A dictionary of properties. Keys must be strings and values must be a numerical or string NBT type.
        :param extra_blocks: A :class:`Block` instance or iterable of :class:`Block` instances
        """
        assert (isinstance(namespace, str) or namespace is None) and isinstance(
            base_name, str
        ), f"namespace and base_name must be strings {namespace} {base_name}"
        self._namespace = namespace
        self._base_name = base_name
        self._namespaced_name = f"{namespace}:{base_name}"

        self._blockstate = None
        self._snbt_blockstate = None
        self._full_blockstate = None

        if properties is None:
            properties = {}
        assert isinstance(properties, dict) and all(
            isinstance(val, PropertyDataTypes) for val in properties.values()
        ), properties

        self._properties = properties
        self._extra_blocks = ()
        if extra_blocks:
            eb = []

            def unpack_block(block_: Iterable[Block]):
                for b in block_:
                    if b.extra_blocks:
                        unpack_block(b)
                    else:
                        eb.append(b)

            unpack_block(extra_blocks)
            self._extra_blocks = tuple(eb)

    @classmethod
    def join(cls, blocks: Iterable[Block]):
        """
        Join an interable of Block objects back into one block class.
        This is the reverse of the __iter__ method.

        >>> stone = Block.from_string_blockstate("minecraft:stone")
        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")
        >>> stone_water = Block.join([stone, water])
        >>> stone_water2 = Block.join(stone_water)
        >>> assert stone_water == stone_water2

        :param blocks: An iterable of one or more Block objects.
        :return: A new Block instance
        """
        blocks = list(blocks)
        return Block(
            blocks[0].namespace, blocks[0].base_name, blocks[0].properties, blocks[1:]
        )

    @classmethod
    def from_string_blockstate(cls, blockstate: str):
        """
        Parse a Java format blockstate where values are all strings and populate a :class:`Block` class with the data.

        >>> stone = Block.from_string_blockstate("minecraft:stone")
        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")

        :param blockstate: The Java blockstate string to parse.
        :return: A Block instance containing the state.
        """
        namespace, block_name, properties = cls.parse_blockstate_string(blockstate)
        return cls(namespace, block_name, properties)

    @classmethod
    def from_snbt_blockstate(cls, blockstate: str):
        """
        Parse a blockstate where values are SNBT of any type and populate a :class:`Block` class with the data.
        """
        namespace, block_name, properties = cls.parse_blockstate_string(
            blockstate, True
        )
        return cls(namespace, block_name, properties)

    @property
    def namespaced_name(self) -> str:
        """
        The namespace:base_name of the blockstate represented by the :class:`Block` object.

        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")
        >>> water.namespaced_name
        'minecraft:water'

        :return: The namespace:base_name of the blockstate
        """
        return self._namespaced_name

    @property
    def namespace(self) -> str:
        """
        The namespace of the blockstate represented by the :class:`Block` object.

        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")
        >>> water.namespace
        'minecraft'

        :return: The namespace of the blockstate
        """
        return self._namespace

    @property
    def base_name(self) -> str:
        """
        The base name of the blockstate represented by the :class:`Block` object.

        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")
        >>> water.base_name
        'water'

        :return: The base name of the blockstate
        """
        return self._base_name

    @property
    def properties(self) -> PropertyType:
        """
        The mapping of properties of the blockstate represented by the :class:`Block` object.

        >>> water = Block.from_string_blockstate("minecraft:water[level=0]")
        >>> water.properties
        {"level": StringTag("0")}

        :return: A dictionary of the properties of the blockstate
        """
        return dict(self._properties)

    @property
    def blockstate(self) -> str:
        """
        The Java blockstate string of this :class:`Block` object
        Note if there are extra blocks this will only show the base block.
        Note this will only contain properties with StringTag values.

        >>> stone = Block("minecraft", "stone")
        >>> stone.blockstate
        'minecraft:stone'
        >>> water = Block("minecraft", "water", {"level": StringTag("0")})
        >>> water.blockstate
        `minecraft:water[level=0]`

        :return: The blockstate string
        """
        if self._blockstate is None:
            self._blockstate = self.namespaced_name
            if self.properties:
                props = [
                    f"{key}={value.py_str}"
                    for key, value in sorted(self.properties.items())
                    if isinstance(value, StringTag)
                ]
                self._blockstate += f"[{','.join(props)}]"
        return self._blockstate

    @property
    def snbt_blockstate(self) -> str:
        """
        A modified version of the Java blockstate format that supports all NBT types.
        Converts the property values to the SNBT format to preserve type.
        Note if there are extra blocks this will only show the base block.

        >>> bell = Block("minecraft", "bell", {
        >>>     "attachment":StringTag("standing"),
        >>>     "direction":IntTag(0),
        >>>     "toggle_bit":ByteTag(0)
        >>> })
        >>> bell.snbt_blockstate
        'minecraft:bell[attachment="standing",direction=0,toggle_bit=0b]'

        :return: The SNBT blockstate string
        """
        if self._snbt_blockstate is None:
            self._snbt_blockstate = self.namespaced_name
            if self.properties:
                props = [
                    f"{key}={value.to_snbt()}"
                    for key, value in sorted(self.properties.items())
                ]
                self._snbt_blockstate += f"[{','.join(props)}]"
        return self._snbt_blockstate

    @property
    def full_blockstate(self) -> str:
        """
        The SNBT blockstate string of the base block and extra blocks.

        >>> bell = Block("minecraft", "bell", {
        >>>     "attachment":StringTag("standing"),
        >>>     "direction":IntTag(0),
        >>>     "toggle_bit":ByteTag(0)
        >>> })
        >>> water = Block("minecraft", "water", {"liquid_depth": IntTag(0)})
        >>> waterlogged_bell = bell + water
        >>> waterlogged_bell.full_blockstate
        'minecraft:bell[attachment="standing",direction=0,toggle_bit=0b]{minecraft:water[liquid_depth=0]}'

        :return: The blockstate string
        """
        if self._full_blockstate is None:
            if self.extra_blocks:
                self._full_blockstate = f"{self.snbt_blockstate}{{{' , '.join(block.snbt_blockstate for block in self.extra_blocks)}}}"
            else:
                self._full_blockstate = self.snbt_blockstate
        return self._full_blockstate

    @property
    def base_block(self) -> Block:
        """
        Returns an instance of :class:`Block` containing only the base block without any extra blocks

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> waterlogged_stone.base_block
        Block(minecraft:stone)

        :return: A Block object
        """
        if self.extra_blocks:
            return Block(
                namespace=self.namespace,
                base_name=self.base_name,
                properties=self.properties,
            )
        else:
            return self

    @property
    def extra_blocks(self) -> Tuple[Block, ...]:
        """
        Returns a tuple of the extra blocks contained in the :class:`Block` instance

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> waterlogged_stone.extra_blocks
        (Block(minecraft:water[level="0"]),)

        :return: A tuple of :class:`Block` objects
        """
        return self._extra_blocks

    @property
    def block_tuple(self) -> Tuple[Block, ...]:
        """
        Returns the stack of blocks represented by this object as a tuple.
        This is a tuple of base_block and extra_blocks

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> waterlogged_stone.block_tuple
        (Block(minecraft:stone), Block(minecraft:water[level="0"]))

        :return: A tuple of :class:`Block` objects
        """
        return (self.base_block,) + self.extra_blocks

    @staticmethod
    def parse_blockstate_string(
        blockstate: str, snbt: bool = False
    ) -> Tuple[str, str, PropertyType]:
        """
        Parse a Java or SNBT blockstate string and return the raw data.

        To parse the blockstate and return a :class:`Block` instance use :func:`from_string_blockstate` or :func:`from_snbt_blockstate`

        :param blockstate: The blockstate to parse
        :param snbt: Are the property values in SNBT format. If false all values must be an instance of :class:`~StringTag`
        :return: namespace, block_name, properties

        """
        if snbt:
            match = Block.snbt_blockstate_regex.match(blockstate)
        else:
            match = Block.blockstate_regex.match(blockstate)
        namespace = match.group("namespace") or "minecraft"
        base_name = match.group("base_name")

        if match.group("property_name") is not None:
            properties = {match.group("property_name"): match.group("property_value")}
            properties_string = match.group("properties")
            if properties_string is not None:
                if snbt:
                    for match in Block.snbt_properties_regex.finditer(
                        properties_string
                    ):
                        properties[match.group("name")] = match.group("value")
                else:
                    for match in Block.properties_regex.finditer(properties_string):
                        properties[match.group("name")] = match.group("value")
        else:
            properties = {}

        if snbt:
            properties_dict = {k: from_snbt(v) for k, v in sorted(properties.items())}
        else:
            properties_dict = {k: StringTag(v) for k, v in sorted(properties.items())}

        return (
            namespace,
            base_name,
            properties_dict,
        )

    def __str__(self) -> str:
        """

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> str(waterlogged_stone)
        'minecraft:stone{minecraft:water[level="0"]}'

        :return: A string showing the information of the :class:`Block` class.
        """
        return self.full_blockstate

    def __repr__(self) -> str:
        """

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> repr(waterlogged_stone)
        'Block(minecraft:stone, extra_blocks=(minecraft:water[level="0"]))'

        :return: The base blockstate string of the Block object along with the blockstate strings of included extra blocks
        """
        if self.extra_blocks:
            return f"Block({self.base_block}, extra_blocks=({', '.join([str(b) for b in self.extra_blocks])}))"
        else:
            return f"Block({self.base_block})"

    def __iter__(self) -> Iterable[Block]:
        """
        Iterate through all the blocks in this :class:`Block` instance.

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> for block in waterlogged_stone:
        >>>     print(block)
        minecraft:stone
        minecraft:water[level="0"]
        """
        yield self.base_block
        yield from self.extra_blocks

    def __len__(self) -> int:
        """
        The number of blocks contained within the :class:`Block` instance.

        >>> waterlogged_stone = Block("minecraft", "stone",
        >>>     extra_blocks=Block("minecraft", "water", {"level": StringTag("0")})
        >>> )
        >>> len(waterlogged_stone)
        2
        """
        return len(self._extra_blocks) + 1

    def __eq__(self, other: Block) -> bool:
        """
        Checks the equality of this Block object to another Block object

        >>> stone = Block("minecraft", "stone")
        >>> stone == stone
        True

        :param other: The Block object to check against
        :return: True if the Blocks objects are equal, False otherwise
        """
        if not isinstance(other, Block):
            return NotImplemented

        return (
            self.namespaced_name == other.namespaced_name
            and self.properties == other.properties
            and self.extra_blocks == other.extra_blocks
        )

    def __gt__(self, other: Block) -> bool:
        """
        Allows blocks to be sorted so numpy.unique can be used on them
        """
        if not isinstance(other, Block):
            return NotImplemented
        return hash(self).__gt__(hash(other))

    def __hash__(self) -> int:
        """
        Hashes the Block object

        :return: A hash of the Block object
        """
        return hash(self.full_blockstate)

    def __add__(self, other: Block) -> Block:
        """
        Add the blocks from `other` to this block.

        >>> stone = Block("minecraft", "stone")
        >>> water = Block("minecraft", "water", {"level": StringTag("0")})
        >>> waterlogged_stone = stone + water
        >>> repr(waterlogged_stone)
        'Block(minecraft:stone, extra_blocks=(minecraft:water[level="0"]))'

        :param other: The :class:`Block` object to add to this :class:`Block`
        :return: A new instance of :class:`Block` with the blocks from other appended to the end of :attr:`extra_blocks`
        """
        if not isinstance(other, Block):
            return NotImplemented

        return Block(
            namespace=self.namespace,
            base_name=self.base_name,
            properties=self.properties,
            extra_blocks=[*self.extra_blocks, other],
        )

    def __sub__(self, other: Block) -> Block:
        """
        Remove all blocks in `other` from the :attr:`extra_blocks` of this instance of :class:`Block`

        >>> stone = Block("minecraft", "stone")
        >>> water = Block("minecraft", "water", {"level": StringTag("0")})
        >>> waterlogged_stone = stone + water
        >>> stone = waterlogged_stone - water

        :param other: The Block object to subtract from this :class:`Block`.
        :return: A new :class:`Block` instance with the blocks in `other` removed from the extra blocks.
        """
        if not isinstance(other, Block):
            return NotImplemented

        extra_blocks_to_remove = set(other)
        new_extras = tuple(
            eb for eb in self.extra_blocks if eb not in extra_blocks_to_remove
        )

        return Block(
            namespace=self.namespace,
            base_name=self.base_name,
            properties=self.properties,
            extra_blocks=new_extras,
        )

    def remove_layer(self, layer: int) -> Block:
        """
        Removes the block at the given index and returns the resulting new Block object.

        >>> stone = Block("minecraft", "stone")
        >>> water = Block("minecraft", "water", {"level": StringTag("0")})
        >>> waterlogged_stone = stone + water
        >>> stone = waterlogged_stone.remove_layer(1)

        :param layer: The layer of extra block to remove.
        :return: A new instance of Block with the same data but with the block at the specified layer removed.
        :raises `BlockException`: Raised when you remove the base block from a Block with no other extra blocks.
        """
        if layer == 0:
            if self.extra_blocks:
                new_base = self._extra_blocks[0]
                return Block(
                    namespace=new_base.namespace,
                    base_name=new_base.base_name,
                    properties=new_base.properties,
                    extra_blocks=[*self._extra_blocks[1:]],
                )
            else:
                raise BlockException(
                    "Removing the base block with no extra blocks is not supported"
                )
        elif layer > len(self.extra_blocks):
            raise BlockException("You cannot remove a non-existent layer")
        else:
            return Block(
                namespace=self.namespace,
                base_name=self.base_name,
                properties=self.properties,
                extra_blocks=[
                    *self.extra_blocks[: layer - 1],
                    *self.extra_blocks[layer:],
                ],
            )

    def __sizeof__(self):
        size = (
            getsizeof(self._namespace)
            + getsizeof(self._base_name)
            + getsizeof(self._namespaced_name)
            + getsizeof(self._properties)
            + getsizeof(self._blockstate)
            + getsizeof(self._extra_blocks)
            + getsizeof(self._snbt_blockstate)
            + getsizeof(self._full_blockstate)
        )
        for eb in self.extra_blocks:
            size += getsizeof(eb)
        return size


# some blocks that probably will not change. Keeping these in one place will make them easier to change if they do.
UniversalAirBlock = Block("universal_minecraft", "air")
# do not rely on this staying the same.
UniversalAirLikeBlocks = (UniversalAirBlock, Block("universal_minecraft", "cave_air"))



from amulet.api.wrapper.format_wrapper import (
    FormatWrapper,
    DefaultSelection,
)
from amulet.api.wrapper.chunk.translator import Translator
from amulet.api.wrapper.chunk.interface import Interface, EntityIDType, EntityCoordType
from amulet.api.wrapper.world_format_wrapper import WorldFormatWrapper
from amulet.api.wrapper.structure_format_wrapper import StructureFormatWrapper



from __future__ import annotations

from .leveldb_39 import (
    LevelDB39Interface as ParentInterface,
)


class LevelDB40Interface(ParentInterface):
    chunk_version = 40

    def __init__(self):
        super().__init__()
        self._set_feature("entities", "actor")


export = LevelDB40Interface



from __future__ import annotations

from .leveldb_40 import (
    LevelDB40Interface as ParentInterface,
)


class LevelDB41Interface(ParentInterface):
    chunk_version = 41

    def __init__(self):
        super().__init__()


export = LevelDB41Interface



from abc import abstractmethod
from typing import Optional

from amulet.api.history.base import RevisionManager
from ..data_types import EntryType


class DBRevisionManager(RevisionManager):
    """A class to hold data about an entries history on disk.
    Revision indexes are still stored in RAM."""

    __slots__ = ("_prefix",)

    def __init__(self, prefix: str, initial_state: EntryType):
        self._prefix: str = prefix
        super().__init__(initial_state)

    def _store_entry(self, entry: EntryType):
        path = f"{self._prefix}/{len(self._revisions)}"
        path = self._serialise(path, entry)
        self._revisions.append(path)

    @abstractmethod
    def _serialise(self, path: str, entry: EntryType) -> Optional[str]:
        raise NotImplementedError

    def get_current_entry(self):
        path = self._revisions[self._current_revision_index]
        return self._deserialise(path)

    @abstractmethod
    def _deserialise(self, path: str) -> EntryType:
        raise NotImplementedError



from __future__ import annotations
from typing import Tuple, Dict, Optional, TYPE_CHECKING

import struct
from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberTuple,
)
from .leveldb_28 import (
    LevelDB28Interface as ParentInterface,
)

if TYPE_CHECKING:
    from amulet.api.chunk.blocks import Blocks


class LevelDB29Interface(ParentInterface):
    chunk_version = 29

    def __init__(self):
        super().__init__()
        self._set_feature("data_2d", "height512|biome4096")

    @staticmethod
    def _chunk_key_to_sub_chunk(cy: int, min_y: int) -> int:
        return cy

    @staticmethod
    def _get_sub_chunk_storage_byte(cy: int, min_y: int) -> bytes:
        return struct.pack("b", cy)


export = LevelDB29Interface



from __future__ import annotations

from typing import NamedTuple, List
import threading
import sys

if sys.version_info >= (3, 10):
    # bisect only supports key as of 3.10
    from bisect import bisect_left, bisect_right

else:

    def bisect_left(a, x, lo=0, hi=None, *, key=None):
        """Return the index where to insert item x in list a, assuming a is sorted.
        The return value i is such that all e in a[:i] have e < x, and all e in
        a[i:] have e >= x.  So if x already appears in the list, a.insert(i, x) will
        insert just before the leftmost x already there.
        Optional args lo (default 0) and hi (default len(a)) bound the
        slice of a to be searched.
        """

        if lo < 0:
            raise ValueError("lo must be non-negative")
        if hi is None:
            hi = len(a)
        # Note, the comparison uses "<" to match the
        # __lt__() logic in list.sort() and in heapq.
        if key is None:
            while lo < hi:
                mid = (lo + hi) // 2
                if a[mid] < x:
                    lo = mid + 1
                else:
                    hi = mid
        else:
            while lo < hi:
                mid = (lo + hi) // 2
                if key(a[mid]) < x:
                    lo = mid + 1
                else:
                    hi = mid
        return lo

    def bisect_right(a, x, lo=0, hi=None, *, key=None):
        """Return the index where to insert item x in list a, assuming a is sorted.
        The return value i is such that all e in a[:i] have e <= x, and all e in
        a[i:] have e > x.  So if x already appears in the list, a.insert(i, x) will
        insert just after the rightmost x already there.
        Optional args lo (default 0) and hi (default len(a)) bound the
        slice of a to be searched.
        """

        if lo < 0:
            raise ValueError("lo must be non-negative")
        if hi is None:
            hi = len(a)
        # Note, the comparison uses "<" to match the
        # __lt__() logic in list.sort() and in heapq.
        if key is None:
            while lo < hi:
                mid = (lo + hi) // 2
                if x < a[mid]:
                    hi = mid
                else:
                    lo = mid + 1
        else:
            while lo < hi:
                mid = (lo + hi) // 2
                if x < key(a[mid]):
                    hi = mid
                else:
                    lo = mid + 1
        return lo


class NoValidSector(Exception):
    """An error for when there is no sector large enough and the region cannot be resized."""

    pass


class Sector(NamedTuple):
    start: int
    stop: int

    @property
    def length(self) -> int:
        return self.stop - self.start

    def intersects(self, other: Sector):
        """Do the two sectors intersect each other."""
        return not (other.stop <= self.start or self.stop <= other.start)

    def contains(self, other: Sector):
        """Is the other sector entirely within this sector."""
        return self.start <= other.start and other.stop <= self.stop

    def neighbours(self, other: Sector):
        """Do the two sectors neighbour but not intersect."""
        return other.stop == self.start or self.stop == other.start

    def split(self, other: Sector) -> List[Sector]:
        """
        Split this sector around another sector.
        The other sector must be contained within this sector

        :param other: The other sector to split around.
        :return: A list of 0-2 sectors
        """
        sectors = []
        if self.start < other.start:
            sectors.append(Sector(self.start, other.start))
        if other.stop < self.stop:
            sectors.append(Sector(other.stop, self.stop))
        return sectors


class SectorManager:
    """A class to manage a sequence of memory."""

    def __init__(self, start: int, stop: int, resizable: bool = True):
        """
        :param start: The start of the memory region
        :param stop: The end of the memory region
        :param resizable: Can the region be resized
        """
        if stop < start:
            raise ValueError("stop must be at least start")
        self._lock = threading.RLock()
        self._stop = stop
        self._resizable = resizable

        # A list of free sectors ordered by start location
        self._free_start = [Sector(start, stop)]
        # A list of free sectors ordered by (length, start).
        # This makes it easier to find the first sector large enough
        self._free_size = [Sector(start, stop)]

        # A set of reserved sectors
        self._reserved = set()

    @property
    def sectors(self) -> List[Sector]:
        """A list of reserved sectors. Ordered by their start location."""
        with self._lock:
            return list(sorted(self._reserved, key=lambda i: i.start))

    def reserve_space(self, length: int) -> Sector:
        """
        Find and reserve a memory location large enough to fit the requested memory.

        :param length: The length of the memory region to reserve
        :return: The index of the start of the reserved memory region
        """
        if not length:
            raise ValueError("Cannot reserve a sector with zero length.")
        with self._lock:
            # find the index of the first element with a larger or equal length prioritising the ones closer to the start
            index = bisect_left(
                self._free_size, (length, 0), key=lambda k: (k.length, k.start)
            )
            if index < len(self._free_size):
                # if there exists a section large enough to fit the length
                free_sector = self._free_size.pop(index)
                start_index = self._free_start.index(free_sector)
                sector = Sector(free_sector.start, free_sector.start + length)
                if free_sector.length > length:
                    free_sector = Sector(sector.stop, free_sector.stop)
                    self._add_size_sector(free_sector)
                    self._free_start[start_index] = free_sector
                else:
                    del self._free_start[start_index]
                self._reserved.add(sector)
                return sector
            elif self._resizable:
                sector = Sector(self._stop, self._stop + length)
                self._stop = sector.stop
                self._reserved.add(sector)
                return sector
            else:
                raise NoValidSector(
                    "There is not enough contiguous space to allocate the length."
                )

    def reserve(self, sector: Sector):
        """
        Mark a section as reserved.
        If you don't know exactly where the sector is use `reserve_space` to find and reserve a new sector

        :param sector: The sector to reserve
        """
        if not sector.length:
            raise ValueError("Cannot reserve a sector with zero length.")
        with self._lock:
            if sector.start >= self._stop and (
                not self._free_start or self._free_start[-1].stop != self._stop
            ):
                if self._resizable:
                    # if the last sector has been reserved or did not exist then create a new one
                    s = Sector(self._stop, sector.stop)
                    self._free_start.append(s)
                    self._add_size_sector(s)
                    self._stop = sector.stop
                else:
                    raise NoValidSector("The sector starts outside of the region.")

            # Get the index of the segment with the latest start point that is
            # less than or equal to the start point of the sector being reserved.
            index = (
                bisect_right(self._free_start, sector.start, key=lambda k: k.start) - 1
            )

            if index < 0:
                raise NoValidSector(
                    "No free sectors that start at or before the one being reserved."
                )

            free_sector = self._free_start[index]

            if sector.stop <= free_sector.stop:
                # The sector fits within the contained sector
                # remove the contained sector from the lists
                del self._free_start[index]
                self._free_size.remove(free_sector)
            elif free_sector.stop == self._stop:
                if self._resizable:
                    # The sector is the last one and the memory region is resizable
                    del self._free_start[index]
                    self._free_size.remove(free_sector)
                    free_sector = Sector(free_sector.start, sector.stop)
                    self._stop = sector.stop
                else:
                    raise NoValidSector(
                        "The sector is outside the defined region and the region is not resizable."
                    )
            else:
                raise NoValidSector("The requested sector is not free to be reserved.")

            # split around the reserved sector
            sectors = free_sector.split(sector)
            # add the results back
            self._free_start[index:index] = sectors
            for s in sectors:
                self._add_size_sector(s)
            self._reserved.add(sector)

    def _add_size_sector(self, sector: Sector):
        self._free_size.insert(
            bisect_left(
                self._free_size,
                (sector.length, sector.start),
                key=lambda k: (k.length, k.start),
            ),
            sector,
        )

    def free(self, sector: Sector):
        """
        Free a reserved sector.
        The sector must match exactly a sector previously reserved.

        :param sector: The sector to free
        """
        with self._lock:
            if sector not in self._reserved:
                raise ValueError("Sector was not reserved")
            # remove the sector from the reserved storage
            self._reserved.remove(sector)

            # Add it back to the free storage
            # find the position where it would be placed in the list ordered by start location
            index = bisect_right(self._free_start, sector.start, key=lambda k: k.start)

            # merge with the right neighbour
            if (
                index < len(self._free_start)
                and self._free_start[index].start == sector.stop
            ):
                right_sector = self._free_start.pop(index)
                self._free_size.remove(right_sector)
                sector = Sector(sector.start, right_sector.stop)

            # merge with the left neighbour
            if index > 0 and self._free_start[index - 1].stop == sector.start:
                left_sector = self._free_start.pop(index - 1)
                self._free_size.remove(left_sector)
                sector = Sector(left_sector.start, sector.stop)
                index -= 1

            self._free_start.insert(index, sector)
            self._add_size_sector(sector)



from __future__ import annotations

from typing import List, TYPE_CHECKING
import logging

from amulet.api.selection import SelectionGroup
from amulet.api.block import Block
from amulet.api.data_types import Dimension

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel

log = logging.getLogger(__name__)


def replace(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup, options: dict
):
    original_blocks = options.get("original_blocks", None)
    if not isinstance(original_blocks, list) and all(
        isinstance(block, Block) for block in original_blocks
    ):
        log.error("Replace operation was not given a list of source Block objects")
        return

    replacement_blocks = options.get("replacement_blocks", None)
    if not isinstance(replacement_blocks, list) and all(
        isinstance(block, Block) for block in replacement_blocks
    ):
        log.error("Replace operation was not given a list of destination Block objects")
        return
    original_blocks: List[Block]
    replacement_blocks: List[Block]

    if len(original_blocks) != len(replacement_blocks):
        if len(replacement_blocks) == 1:
            replacement_blocks = replacement_blocks * len(original_blocks)
        else:
            log.error(
                "Replace operation must be given the same number of destination blocks as source blocks"
            )

    original_internal_ids = list(
        map(world.block_palette.get_add_block, original_blocks)
    )
    replacement_internal_ids = list(
        map(world.block_palette.get_add_block, replacement_blocks)
    )

    for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection):
        old_blocks = chunk.blocks[slices]
        new_blocks = old_blocks.copy()
        for original_id, replacement_id in zip(
            original_internal_ids, replacement_internal_ids
        ):
            new_blocks[old_blocks == original_id] = replacement_id
        chunk.blocks[slices] = new_blocks
        chunk.changed = True



from typing import Tuple, Generator

from amulet.utils.generator import generator_unpacker
from amulet.api.history.base.history_manager import HistoryManager
from .container import ContainerHistoryManager

SnapshotType = Tuple[HistoryManager, ...]


class MetaHistoryManager(ContainerHistoryManager):
    def __init__(self):
        super().__init__()
        self._world_managers = []
        self._non_world_managers = []

    def _check_snapshot(self, snapshot: SnapshotType):
        assert isinstance(snapshot, tuple) and all(
            isinstance(item, HistoryManager) for item in snapshot
        )

    def register(self, manager: HistoryManager, is_world_manager: bool):
        """
        Register a manager to track.
        :param manager: The manager to track
        :param is_world_manager: Is the manager tracking world data
        :return:
        """
        assert isinstance(
            manager, HistoryManager
        ), "manager must be an instance of BaseHistoryManager"
        if is_world_manager:
            self._world_managers.append(manager)
        else:
            self._non_world_managers.append(manager)

    def _undo(self, snapshot: SnapshotType):
        for item in snapshot:
            item.undo()

    def _redo(self, snapshot: SnapshotType):
        for item in snapshot:
            item.redo()

    def _mark_saved(self):
        for manager in self._managers():
            manager.mark_saved()

    def _managers(
        self, world: bool = True, non_world: bool = True
    ) -> Tuple[HistoryManager, ...]:
        return (
            tuple(self._world_managers) * world
            + tuple(self._non_world_managers) * non_world
        )

    @property
    def changed(self) -> bool:
        if super().changed:
            return True
        managers = self._managers(True, False)
        for manager in managers:
            if manager.changed:
                return True
        return False

    def create_undo_point(self, world=True, non_world=True) -> bool:
        return generator_unpacker(self.create_undo_point_iter(world, non_world))

    def create_undo_point_iter(
        self, world=True, non_world=True
    ) -> Generator[float, None, bool]:
        """Create an undo point.

        :param world: Should the world based history managers be included
        :param non_world: Should the non-world based history managers be included
        :return:
        """
        managers = self._managers(world, non_world)
        snapshot = []
        for manager in managers:
            changed = yield from manager.create_undo_point_iter()
            if changed:
                snapshot.append(manager)
        return self._register_snapshot(tuple(snapshot))

    def restore_last_undo_point(self):
        for manager in self._managers():
            manager.restore_last_undo_point()

    def purge(self):
        """Unload all history data. Restore to the state after creation."""
        for manager in self._managers():
            manager.purge()



from __future__ import annotations

import itertools
import numpy

from typing import Tuple, Iterable, List, Union, Optional, overload, Set

from amulet.api.data_types import (
    BlockCoordinates,
    CoordinatesAny,
    ChunkCoordinates,
    SubChunkCoordinates,
    FloatTriplet,
    PointCoordinatesAny,
)
from .abstract_selection import AbstractBaseSelection
from .box import SelectionBox


class SelectionGroup(AbstractBaseSelection):
    """
    A container for zero or more :class:`SelectionBox` instances.

    This allows for non-rectangular and non-contiguous selections.
    """

    def __init__(
        self, selection_boxes: Union[SelectionBox, Iterable[SelectionBox]] = ()
    ):
        """
        Construct a new :class:`SelectionGroup` class from the given data.

        >>> SelectionGroup(SelectionBox((0, 0, 0), (1, 1, 1)))
        >>> SelectionGroup([
        >>>     SelectionBox((0, 0, 0), (1, 1, 1)),
        >>>     SelectionBox((1, 1, 1), (2, 2, 2))
        >>> ])

        :param selection_boxes: A :class:`SelectionBox` or iterable of :class:`SelectionBox` classes.
        """
        if isinstance(selection_boxes, SelectionBox):
            self._selection_boxes: Tuple[SelectionBox, ...] = (selection_boxes,)
        else:
            self._selection_boxes: Tuple[SelectionBox, ...] = tuple(
                box for box in selection_boxes if isinstance(box, SelectionBox)
            )

    def __repr__(self) -> str:
        boxes = ", ".join([repr(box) for box in self.selection_boxes])
        return f"SelectionGroup([{boxes}])"

    def __str__(self) -> str:
        boxes = ", ".join([str(box) for box in self.selection_boxes])
        return f"[{boxes}]"

    def __eq__(self, other: SelectionGroup) -> bool:
        """
        Does the contents of this :class:`SelectionGroup` match the other :class:`SelectionGroup`.

        Note if the boxes do not exactly match this will return False even if the volume represented is the same.

        :param other: The other :class:`SelectionGroup` to compare with.
        :return: True if the boxes contained match.
        """
        return (
            isinstance(other, SelectionGroup)
            and self.selection_boxes_sorted == other.selection_boxes_sorted
        )

    def __add__(self, boxes: Iterable[SelectionBox]) -> SelectionGroup:
        """
        Add an iterable of :class:`SelectionBox` classes to this :class:`SelectionGroup`.

        Note this will construct a new :class:`SelectionGroup` because it is immutable so cannot be modified in place.

        >>> group1 = SelectionGroup(SelectionBox((-1, -1, -1), (0, 0, 0)))
        >>> group2 = SelectionGroup([
        >>>     SelectionBox((0, 0, 0), (1, 1, 1)),
        >>>     SelectionBox((1, 1, 1), (2, 2, 2))
        >>> ])
        >>> group1 + group2
        SelectionGroup([SelectionBox((-1, -1, -1), (0, 0, 0)), SelectionBox((0, 0, 0), (1, 1, 1)), SelectionBox((1, 1, 1), (2, 2, 2))])
        >>> group1 += group2
        >>> group1
        SelectionGroup([SelectionBox((-1, -1, -1), (0, 0, 0)), SelectionBox((0, 0, 0), (1, 1, 1)), SelectionBox((1, 1, 1), (2, 2, 2))])

        :param boxes: An iterable of boxes to add to this group.
        :return: A new :class:`SelectionGroup` class containing the boxes from this instance and those in ``boxes``.
        """
        try:
            boxes = tuple(boxes)
        except:
            raise NotImplemented
        if all(isinstance(b, SelectionBox) for b in boxes):
            return SelectionGroup(tuple(self) + boxes)
        raise NotImplemented

    def __iter__(self) -> Iterable[SelectionBox]:
        """An iterable of all the :class:`SelectionBox` classes in the group."""
        yield from self._selection_boxes

    def __len__(self) -> int:
        """The number of :class:`SelectionBox` classes in the group."""
        return len(self._selection_boxes)

    def __contains__(self, item: CoordinatesAny) -> bool:
        return self.contains_block(item)

    def contains_block(self, coords: CoordinatesAny) -> bool:
        return any(box.contains_block(coords) for box in self._selection_boxes)

    def contains_point(self, coords: CoordinatesAny) -> bool:
        return any(box.contains_point(coords) for box in self._selection_boxes)

    @property
    def blocks(self) -> Iterable[BlockCoordinates]:
        """
        The location of every block in the selection.

        >>> for x, y, z in group.blocks:
        >>>     ...

        Note: if boxes intersect, the blocks in the intersected region will be included multiple times.

        If this behaviour is not desired the :meth:`merge_boxes` method will return a new SelectionGroup with no intersections.

        >>> for x, y, z in group.merge_boxes().blocks:
        >>>     ...

        :return: An iterable of block locations.
        """
        return itertools.chain.from_iterable(self.selection_boxes)

    def __bool__(self) -> bool:
        """Are there any boxes in the group."""
        return bool(self._selection_boxes)

    @overload
    def __getitem__(self, item: int) -> SelectionBox: ...

    @overload
    def __getitem__(self, item: slice) -> SelectionGroup: ...

    def __getitem__(self, item):
        """Get the selection box at the given index."""
        val = self._selection_boxes[item]
        if isinstance(val, tuple):
            return SelectionGroup(val)
        else:
            return val

    @property
    def min(self) -> BlockCoordinates:
        """The minimum point of of all the boxes in the group."""
        return tuple(self.min_array.tolist())

    @property
    def min_array(self) -> numpy.ndarray:
        """The minimum point of of all the boxes in the group as a numpy array."""
        if self._selection_boxes:
            return numpy.min(numpy.array([box.min for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SelectionBoxes")

    @property
    def min_x(self) -> int:
        return int(self.min_array[0])

    @property
    def min_y(self) -> int:
        return int(self.min_array[1])

    @property
    def min_z(self) -> int:
        return int(self.min_array[2])

    @property
    def max(self) -> BlockCoordinates:
        """The maximum point of of all the boxes in the group."""
        return tuple(self.max_array.tolist())

    @property
    def max_array(self) -> numpy.ndarray:
        """The maximum point of of all the boxes in the group as a numpy array."""
        if self._selection_boxes:
            return numpy.max(numpy.array([box.max for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SelectionBoxes")

    @property
    def max_x(self) -> int:
        return int(self.max_array[0])

    @property
    def max_y(self) -> int:
        return int(self.max_array[1])

    @property
    def max_z(self) -> int:
        return int(self.max_array[2])

    @property
    def bounds(self) -> Tuple[BlockCoordinates, BlockCoordinates]:
        return self.min, self.max

    @property
    def bounds_array(self) -> numpy.ndarray:
        return numpy.array([self.min_array, self.max_array])

    def to_box(self) -> SelectionBox:
        """Create a `SelectionBox` based off the bounds of the boxes in the group."""
        return SelectionBox(self.min, self.max)

    def merge_boxes(self) -> SelectionGroup:
        """
        Take the boxes as they were given to this class, merge neighbouring boxes and remove overlapping regions.

        The result should be a SelectionGroup containing one or more SelectionBox classes that represents the same
        volume as the original but with no overlapping boxes.
        """
        selection_boxes = self.selection_boxes
        # TODO remove duplicate boxes

        # remove duplicates
        selection_boxes_ = []
        for box in selection_boxes:
            if not any(box == box_ for box_ in selection_boxes_):
                selection_boxes_.append(box)
        selection_boxes = selection_boxes_

        if len(selection_boxes) >= 2:
            merge_boxes = True
            while merge_boxes:
                # find two neighbouring boxes and merge them
                merge_boxes = False  # if two boxes get merged this will be set back to True and this will run again.
                box_index = 0  # the index of the first box
                while box_index < len(selection_boxes):
                    box = selection_boxes[box_index]
                    other_index = box_index + 1  # the index of the second box.
                    # This always starts at one greater than box_index because
                    # the lower values were already checked the other way around
                    while other_index < len(selection_boxes):
                        other = selection_boxes[other_index]
                        x_dim = box.min_x == other.min_x and box.max_x == other.max_x
                        y_dim = box.min_y == other.min_y and box.max_y == other.max_y
                        z_dim = box.min_z == other.min_z and box.max_z == other.max_z

                        x_border = box.max_x == other.min_x or other.max_x == box.min_x
                        y_border = box.max_y == other.min_y or other.max_y == box.min_y
                        z_border = box.max_z == other.min_z or other.max_z == box.min_z

                        if (
                            (x_dim and y_dim and z_border)
                            or (x_dim and z_dim and y_border)
                            or (y_dim and z_dim and x_border)
                        ):
                            selection_boxes.pop(other_index)
                            selection_boxes.pop(box_index)
                            selection_boxes.append(
                                SelectionBox(
                                    numpy.min([box.min, other.min], 0),
                                    numpy.max([box.max, other.max], 0),
                                )
                            )
                            merge_boxes = True
                            box = selection_boxes[box_index]
                            other_index = box_index + 1
                        else:
                            other_index += 1
                    box_index += 1
        return SelectionGroup(selection_boxes)

    @property
    def is_contiguous(self) -> bool:
        """
        Does the SelectionGroup represent one connected region (True) or multiple separated regions (False).

        If two boxes are touching at the corners this is classed as contiguous.
        """
        # TODO: This needs some work. It will only work if the selections touch in a chain
        #  it does not care if it loops back and intersects itself. Does intersecting count as being contiguous?
        #  I would say yes
        if len(self._selection_boxes) == 1:
            return True

        for i in range(len(self._selection_boxes) - 1):
            sub_box: SelectionBox = self._selection_boxes[i]
            next_box: SelectionBox = self._selection_boxes[i + 1]
            if not sub_box.touches(next_box):
                return False

        return True

    @property
    def is_rectangular(self) -> bool:
        """
        Checks if the SelectionGroup is a rectangle

        :return: True is the selection is a rectangle, False otherwise
        """
        return (
            len(self._selection_boxes) == 1
            or len(self.merge_boxes().selection_boxes) == 1
        )

    @property
    def selection_boxes(self) -> Tuple[SelectionBox, ...]:
        """
        A tuple of the :class:`SelectionBox` instances stored for this group.
        """
        return self._selection_boxes

    @property
    def selection_boxes_sorted(self) -> List[SelectionBox]:
        """
        A list of the :class:`SelectionBox` instances for this group sorted by their hash.
        """
        return sorted(self._selection_boxes, key=hash)

    def chunk_count(self, sub_chunk_size: int = 16) -> int:
        return len(self.chunk_locations(sub_chunk_size))

    def chunk_locations(self, sub_chunk_size: int = 16) -> Set[ChunkCoordinates]:
        return set(
            location
            for box in self.selection_boxes
            for location in box.chunk_locations(sub_chunk_size)
        )

    def chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[ChunkCoordinates, SelectionBox]]:
        for box in self.selection_boxes:
            yield from box.chunk_boxes(sub_chunk_size)

    def sub_chunk_count(self, sub_chunk_size: int = 16) -> int:
        return len(self.sub_chunk_locations(sub_chunk_size))

    def sub_chunk_locations(self, sub_chunk_size: int = 16) -> Set[SubChunkCoordinates]:
        yield from set(
            location
            for box in self.selection_boxes
            for location in box.sub_chunk_locations(sub_chunk_size)
        )

    def sub_chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Iterable[Tuple[SubChunkCoordinates, SelectionBox]]:
        for box in self.selection_boxes:
            for (cx, cy, cz), sub_box in box.sub_chunk_boxes(sub_chunk_size):
                yield (cx, cy, cz), box

    def intersects(self, other: Union[SelectionGroup, SelectionBox]) -> bool:
        if isinstance(other, SelectionGroup):
            return any(
                self_box.intersects(other_box)
                for self_box in self.selection_boxes
                for other_box in other.selection_boxes
            )
        elif isinstance(other, SelectionBox):
            return any(self_box.intersects(other) for self_box in self.selection_boxes)

    @staticmethod
    def _to_group(other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        if not isinstance(other, (SelectionGroup, SelectionBox)):
            raise TypeError("other must be a SelectionGroup or SelectionBox.")
        if isinstance(other, SelectionBox):
            other = SelectionGroup(other)
        return other

    def intersection(
        self, other: Union[SelectionGroup, SelectionBox]
    ) -> SelectionGroup:
        other = self._to_group(other)
        intersection = []
        for self_box in self.selection_boxes:
            for other_box in other.selection_boxes:
                if self_box.intersects(other_box):
                    intersection.append(self_box.intersection(other_box))
        return SelectionGroup(intersection)

    def subtract(self, other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        """
        Returns a new :class:`SelectionGroup` containing the volume that does not intersect with other.

        This may be empty if other fully contains self or equal to self if they do not intersect.

        :param other: The :class:`SelectionBox` or :class:`SelectionGroup` to subtract.
        """
        other = self._to_group(other)
        selections = self
        for other_box in other:
            # for each box in other
            selections_new = SelectionGroup()
            for self_box in selections:
                selections_new += self_box.subtract(other_box)
            selections = selections_new
            if not selections:
                break
        return selections

    def union(self, other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        """
        Returns a new SelectionGroup containing the volume of self and other.

        :param other: The other selection to add to this one.
        """
        other = self._to_group(other)
        if other.is_subset(self):
            return self
        else:
            return self.subtract(other) + other

    def is_subset(self, other: Union[SelectionGroup, SelectionBox]) -> bool:
        """
        Is this selection completely contained within ``other``.

        :param other: The other selection to test against.
        :return: True if this selection completely fits in other.
        """
        other = self._to_group(other)
        return not self.subtract(other)

    def closest_vector_intersection(
        self, origin: PointCoordinatesAny, vector: PointCoordinatesAny
    ) -> Tuple[Optional[int], float]:
        """
        Returns the index for the closest box in the look vector and the multiplier of the look vector to get there.

        :param origin: The origin tuple of the vector
        :param vector: The vector magnitude in x, y and z
        :return: Index for the closest box and the multiplier of the vector to get there. None, inf if no intersection.
        """
        index_return = None
        multiplier = float("inf")
        for index, box in enumerate(self._selection_boxes):
            mult = box.intersects_vector(origin, vector)
            if mult is not None and mult < multiplier:
                multiplier = mult
                index_return = index
        return index_return, multiplier

    def transform(
        self, scale: FloatTriplet, rotation: FloatTriplet, translation: FloatTriplet
    ) -> SelectionGroup:
        """
        Creates a new :class:`SelectionGroup` transformed by the given inputs.

        :param scale: A tuple of scaling factors in the x, y and z axis.
        :param rotation: The rotation about the x, y and z axis in radians.
        :param translation: The translation about the x, y and z axis.
        :return: A new :class:`~amulet.api.selection.SelectionGroup` representing the transformed selection.
        """
        selection_group = SelectionGroup()
        for selection in self.selection_boxes:
            selection_group += selection.transform(scale, rotation, translation)
        return selection_group

    @property
    def volume(self) -> int:
        return sum(box.volume for box in self.selection_boxes)

    @property
    def footprint_area(self) -> int:
        """
        The 2D area that the selection fills when looking at the selection from above.
        """
        return (
            SelectionGroup(
                [
                    SelectionBox((box.min_x, 0, box.min_z), (box.max_x, 1, box.max_z))
                    for box in self.selection_boxes
                ]
            )
            .merge_boxes()
            .volume
        )


if __name__ == "__main__":
    b1 = SelectionBox((0, 0, 0), (4, 4, 4))
    b2 = SelectionBox((7, 7, 7), (10, 10, 10))
    sel = SelectionGroup((b1, b2))

    for x, y, z in sel.blocks:
        print(x, y, z)



from __future__ import annotations

from .leveldb_7 import (
    LevelDB7Interface,
)

from typing import Tuple, Dict, Optional, TYPE_CHECKING, List

import struct
import numpy
from numpy.typing import NDArray
from amulet_nbt import NamedTag, CompoundTag, StringTag, IntTag, ShortTag

from amulet.api.block import Block, PropertyDataTypes

from amulet.utils.numpy_helpers import brute_sort_objects_no_hash
from amulet.api.data_types import (
    AnyNDArray,
    VersionIdentifierTuple,
)

if TYPE_CHECKING:
    from amulet.api.chunk.blocks import Blocks


PackedBlockT = Tuple[Tuple[Optional[int], Block], ...]


class LevelDB8Interface(LevelDB7Interface):
    chunk_version = 8

    def __init__(self):
        super().__init__()
        self._set_feature("terrain", "2fnpalette")

    def _encode_subchunks(
        self,
        blocks: "Blocks",
        palette: AnyNDArray,
        bounds: Tuple[int, int],
        max_world_version: VersionIdentifierTuple,
    ) -> Dict[int, Optional[bytes]]:
        # Encode sub-chunk block format 8
        # TODO: untangle this mess. The lack of typing in numpy is just making this harder.
        palette_list: List[PackedBlockT] = list(palette)
        min_y = bounds[0] // 16
        max_y = bounds[1] // 16
        if palette_list:
            if palette_list[0][0][0] is None:
                air = NamedTag(
                    CompoundTag(
                        {
                            "name": StringTag("minecraft:air"),
                            "val": ShortTag(0),
                        }
                    )
                )
            else:
                air = NamedTag(
                    CompoundTag(
                        {
                            "name": StringTag("minecraft:air"),
                            "states": CompoundTag({}),
                            "version": IntTag(17_629_184),  # 1, 13, 0, 0
                        }
                    )
                )

            packed_palette: List[Tuple[NamedTag, ...]] = []
            for index, block in enumerate(palette_list):
                full_block: List[NamedTag] = []
                for sub_block_version, sub_block in block:
                    properties = sub_block.properties
                    if sub_block_version is None:
                        block_data = properties.get("block_data", IntTag(0))
                        if isinstance(block_data, IntTag):
                            block_data = block_data.py_int
                            # if block_data >= 16:
                            #     block_data = 0
                        else:
                            block_data = 0
                        sub_block_ = NamedTag(
                            CompoundTag(
                                {
                                    "name": StringTag(sub_block.namespaced_name),
                                    "val": ShortTag(block_data),
                                }
                            )
                        )
                    else:
                        sub_block_ = NamedTag(
                            CompoundTag(
                                {
                                    "name": StringTag(sub_block.namespaced_name),
                                    "states": CompoundTag(
                                        {
                                            key: val
                                            for key, val in properties.items()
                                            if isinstance(val, PropertyDataTypes)
                                        }
                                    ),
                                    "version": IntTag(sub_block_version),
                                }
                            )
                        )

                    full_block.append(sub_block_)
                packed_palette.append(tuple(full_block))

            chunk = {}
            palette_depth = numpy.array([len(block) for block in packed_palette])
            for cy in range(min_y, max_y):
                if cy in blocks:
                    palette_index, sub_chunk = numpy.unique(
                        blocks.get_sub_chunk(cy), return_inverse=True
                    )
                    sub_chunk_palette: List[Tuple[NamedTag, ...]] = [
                        packed_palette[i] for i in palette_index
                    ]
                    sub_chunk_depth = palette_depth[palette_index].max()

                    if (
                        sub_chunk_depth == 1
                        and len(sub_chunk_palette) == 1
                        and sub_chunk_palette[0][0].compound.get_string("name").py_str
                        == "minecraft:air"
                    ):
                        chunk[cy] = None
                    else:
                        # pad block_palette with air in the extra layers
                        sub_chunk_palette_full: NDArray[NamedTag] = numpy.empty(
                            (len(sub_chunk_palette), sub_chunk_depth), dtype=object
                        )
                        sub_chunk_palette_full.fill(air)

                        for index, block_tuple in enumerate(sub_chunk_palette):
                            for sub_index, block in enumerate(block_tuple):
                                sub_chunk_palette_full[index, sub_index] = block
                        # should now be a 2D array with an NamedTag in each element

                        if max_world_version[1] >= (
                            1,
                            17,
                            30,
                        ):  # Why do I need to check against game version and not chunk version
                            sub_chunk_bytes = [
                                b"\x09",
                                bytes([sub_chunk_depth]),
                                struct.pack("b", cy),
                            ]
                        else:
                            sub_chunk_bytes = [b"\x08", bytes([sub_chunk_depth])]
                        for sub_chunk_layer_index in range(sub_chunk_depth):
                            # TODO: sort out a way to do this quicker without brute forcing it.
                            (
                                sub_chunk_layer_palette,
                                sub_chunk_remap,
                            ) = brute_sort_objects_no_hash(
                                sub_chunk_palette_full[:, sub_chunk_layer_index]
                            )
                            sub_chunk_layer = sub_chunk_remap[sub_chunk.ravel()]

                            # sub_chunk_layer, sub_chunk_layer_palette = sub_chunk, sub_chunk_palette_full[:, sub_chunk_layer_index]
                            sub_chunk_bytes.append(
                                self._save_palette_subchunk(
                                    sub_chunk_layer.reshape(16, 16, 16),
                                    list(sub_chunk_layer_palette.ravel()),
                                )
                            )

                        chunk[cy] = b"".join(sub_chunk_bytes)
                else:
                    chunk[cy] = None
        else:
            chunk = {i: None for i in range(min_y, max_y)}

        return chunk


export = LevelDB8Interface



from typing import Dict, overload, List, Union, Tuple, Generator, Iterable
from numpy import character, integer

from amulet_nbt import ByteTag, IntTag, ShortTag, LongTag, StringTag
from amulet.api.data_types import Int, BiomeType
from amulet.api.registry.base_registry import BaseRegistry


class BiomeManager(BaseRegistry):
    """
    Class to handle the mappings between biome strings and their index-based internal IDs
    """

    def __init__(self, biomes: Iterable[BiomeType] = ()):
        """
        Creates a new BiomeManager object
        """
        self._index_to_biome: List[BiomeType] = []
        self._biome_to_index: Dict[BiomeType, int] = {}
        for biome in biomes:
            # if a list is given it is assumed that the biome palette will be the same size as the list.
            # Ensure that if a value is duplicated it will appear twice in the list
            assert isinstance(biome, str), f"biome must be a string. Got {biome}"
            if biome not in self._biome_to_index:
                self._biome_to_index[biome] = len(self._index_to_biome)
            self._index_to_biome.append(biome)

    def __len__(self):
        """
        The number of biomes in the registry.

        >>> len(level.biome_palette)
        10
        """
        return len(self._index_to_biome)

    def __contains__(self, item: Union[int, BiomeType]):
        """
        Is the given biome string already in the registry.

        >>> biome_string in level.biome_palette
        True
        >>> 7 in level.biome_palette
        True

        :param item: The biome or id to check.
        """
        if isinstance(item, int):
            return item < len(self._index_to_biome)
        elif isinstance(item, str):
            return item in self._biome_to_index
        return False

    def __iter__(self) -> Iterable[BiomeType]:
        """
        Iterate through all biomes in the registry.

        >>> for biome in level.biome_palette:
        >>>     ...
        """
        yield from self._index_to_biome

    @property
    def biomes(self) -> Tuple[BiomeType]:
        """
        The biomes in the registry as a tuple.
        """
        return tuple(self._index_to_biome)

    def values(self) -> Tuple[BiomeType]:
        """
        The biomes in the registry as a tuple.
        """
        return self.biomes

    def items(self) -> Generator[Tuple[int, BiomeType], None, None]:
        """
        A generator of the biome indexes and the biome strings.
        """
        yield from enumerate(self._index_to_biome)

    @overload
    def __getitem__(self, item: BiomeType) -> int: ...

    @overload
    def __getitem__(self, item: Int) -> BiomeType: ...

    @overload
    def __getitem__(
        self, item: Iterable[Union[Int, BiomeType]]
    ) -> List[Union[BiomeType, Int]]: ...

    def __getitem__(self, item):
        """
        If a string is passed to this function, it will return the internal ID/index of the biome.

        If an int is given, this method will return the biome string at that specified index.

        >>> level.biome_palette[biome]
        7
        >>> level.biome_palette[7]
        biome

        :param item: The string or int to get the mapping data of
        :return: An int if a string was supplied, a string if an int was supplied
        :raises KeyError if the requested item is not present.
        """
        try:
            return self._get_item(item)
        except (KeyError, IndexError):
            raise KeyError(
                f"There is no {item} in the BiomeManager. "
                f"You might want to use the `get_add_biome` function for the biome before accessing them."
            )

    def _get_item(self, item):
        if isinstance(item, (str, character, StringTag)):
            return self._biome_to_index[str(item)]
        elif isinstance(item, (int, integer, ByteTag, ShortTag, IntTag, LongTag)):
            return self._index_to_biome[int(item)]
        # if it isn't an int or string assume an iterable of the above.
        return [self._get_item(i) for i in item]

    def get_add_biome(self, biome: BiomeType) -> int:
        """
        Adds a biome string to the internal biome string/ID mappings.

        If the biome already exists in the mappings, the existing ID is returned.

        :param biome: The biome string to add to the manager
        :return: The internal ID of the biome
        """
        assert isinstance(biome, str), f"biome must be a string. Got {biome}"
        if biome not in self._biome_to_index:
            self._biome_to_index[biome] = len(self._index_to_biome)
            self._index_to_biome.append(biome)
        return self._biome_to_index[biome]

    def register(self, biome: BiomeType) -> int:
        """
        An alias of :meth:`get_add_biome`.

        Adds a biome string to the internal biome string/ID mappings.

        If the biome already exists in the mappings, the existing ID is returned.

        :param biome: The biome string to add to the manager
        :return: The internal ID of the biome
        """
        return self.get_add_biome(biome)



from __future__ import annotations

from .anvil_1467 import Anvil1467Interface as ParentInterface


class Anvil1484Interface(ParentInterface):
    """
    Changed height keys
    """

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "C|V2")

    @staticmethod
    def minor_is_valid(key: int):
        return 1484 <= key < 1503


export = Anvil1484Interface



from __future__ import annotations

from .anvil_2681 import (
    Anvil2681Interface as ParentInterface,
)


class Anvil2709Interface(ParentInterface):
    """
    Made height bit depth variable to store increased heights
    Made the biome array size variable to handle the increased height
    """

    @staticmethod
    def minor_is_valid(key: int):
        return 2709 <= key < 2800


export = Anvil2709Interface



from typing import TYPE_CHECKING, Any, Tuple
import numpy

from amulet.api.wrapper import Interface, EntityIDType, EntityCoordType
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox
from amulet.api.data_types import AnyNDArray, VersionIdentifierType, VersionNumberAny
from amulet.level.loader import Translators
from amulet.api.block import Block
from .chunk import SpongeSchemChunk

from amulet_nbt import NamedTag

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from PyMCTranslate import TranslationManager


class SpongeSchemInterface(Interface):
    _entity_id_type = EntityIDType.namespace_str_Id
    _entity_coord_type = EntityCoordType.Pos_list_double
    _block_entity_id_type = EntityIDType.namespace_str_Id
    _block_entity_coord_type = EntityCoordType.Pos_array_int

    def is_valid(self, key: Tuple) -> bool:
        return True

    def decode(
        self, cx: int, cz: int, data: SpongeSchemChunk
    ) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format
        :param cx: chunk x coordinate
        :param cz: chunk z coordinate
        :param data: Raw chunk data provided by the format.
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        """
        palette = numpy.empty(len(data.palette) + 1, dtype=object)
        palette[0] = Block(namespace="minecraft", base_name="air")
        palette[1:] = data.palette[:]

        chunk = Chunk(cx, cz)
        box = data.selection.create_moved_box((cx * 16, 0, cz * 16), subtract=True)
        chunk.blocks[box.slice] = data.blocks + 1
        for tag in data.block_entities:
            block_entity = self._decode_block_entity(
                NamedTag(tag), self._block_entity_id_type, self._block_entity_coord_type
            )
            if block_entity is not None:
                chunk.block_entities.insert(block_entity)
        for tag in data.entities:
            entity = self._decode_entity(
                NamedTag(tag), self._block_entity_id_type, self._block_entity_coord_type
            )
            if entity is not None:
                chunk.entities.append(entity)

        return chunk, palette

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: VersionIdentifierType,
        box: SelectionBox,
    ) -> SpongeSchemChunk:
        """
        Take a version-specific chunk and encode it to raw data for the format to store.
        :param chunk: The already translated version-specfic chunk to encode.
        :param palette: The block_palette the ids in the chunk correspond to.
        :type palette: numpy.ndarray[Block]
        :param max_world_version: The key to use to find the encoder.
        :param box: The volume of the chunk to pack.
        :return: Raw data to be stored by the Format.
        """
        entities = []
        for entity in chunk.entities:
            if entity.location in box:
                entities.append(
                    self._encode_entity(
                        entity, self._entity_id_type, self._entity_coord_type
                    ).compound
                )
        block_entities = []
        for block_entity in chunk.block_entities:
            if block_entity.location in box:
                block_entities.append(
                    self._encode_block_entity(
                        block_entity,
                        self._block_entity_id_type,
                        self._block_entity_coord_type,
                    ).compound
                )

        slices = box.create_moved_box(
            (chunk.cx * 16, 0, chunk.cz * 16), subtract=True
        ).slice

        return SpongeSchemChunk(
            box,
            numpy.asarray(chunk.blocks[slices]),
            palette.copy(),
            block_entities,
            entities,
        )

    def get_translator(
        self,
        max_world_version: Tuple[str, int],
        data: Any = None,
        translation_manager: "TranslationManager" = None,
    ) -> Tuple["Translator", VersionNumberAny]:
        platform, version_number = max_world_version
        if platform != "java":
            raise ValueError("Platform must be java")
        version = translation_manager.get_version(platform, version_number)
        version_number = version.data_version
        return Translators.get((platform, version_number)), version_number



from __future__ import annotations
from typing import Generator

from ..base_level import BaseLevel
from .void_format_wrapper import VoidFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.data_types import Dimension
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.utils.generator import generator_unpacker
import copy


class ImmutableStructure(BaseLevel):
    """
    This is a special version of the level class that is only used when extracting a region of the world.

    It is much the same as a normal level class but does not have an associated file/folder data like :class:`~amulet.api.level.World` and :class:`~amulet.api.level.Structure` do.

    To extract a section of a world you should use :meth:`~amulet.api.level.BaseLevel.extract_structure`.
    """

    def __init__(
        self,
    ):
        """
        Construct an :class:`ImmutableStructure` instance.

        You probably don't want to call this directly.

        To extract a section of a world you should use :meth:`~amulet.api.level.BaseLevel.extract_structure`.
        """
        super().__init__("", VoidFormatWrapper(""))
        self._selection = SelectionGroup(
            [
                SelectionBox(
                    (-30_000_000, -30_000_000, -30_000_000),
                    (30_000_000, 30_000_000, 30_000_000),
                )
            ]
        )

    @property
    def selection_bounds(self) -> SelectionGroup:
        return self._selection

    def bounds(self, dimension: Dimension) -> SelectionGroup:
        return self._selection

    def undo(self):
        pass

    def redo(self):
        pass

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        super().put_chunk(chunk, dimension)
        self.history_manager.create_undo_point()

    @classmethod
    def from_level(
        cls, level: BaseLevel, selection: SelectionGroup, dimension: Dimension
    ):
        """
        Extract a section of the level into an :class:`ImmutableStructure` class.

        :param level: The level to extract the area from.
        :param selection: The selection to extract.
        :param dimension: The dimension to extract from.
        :return: The created instance of :class:`ImmutableStructure`
        """
        return generator_unpacker(cls.from_level_iter(level, selection, dimension))

    @classmethod
    def from_level_iter(
        cls, level: BaseLevel, selection: SelectionGroup, dimension: Dimension
    ) -> Generator[float, None, ImmutableStructure]:
        """
        Extract a section of the level into an :class:`ImmutableStructure` class.

        Also yields the progress from 0-1.

        :param level: The level to extract the area from.
        :param selection: The selection to extract.
        :param dimension: The dimension to extract from.
        :return: The created instance of :class:`ImmutableStructure`
        """
        self = cls()
        self._selection = selection
        dst_dimension = self.dimensions[0]
        count = len(list(level.get_coord_box(dimension, selection)))
        for index, (chunk, _) in enumerate(level.get_chunk_boxes(dimension, selection)):
            self.put_chunk(copy.deepcopy(chunk), dst_dimension)
            yield (index + 1) / count
        return self



from typing import Union, Tuple, overload, Iterable, Optional, Dict  # , Literal
import numpy
import math

from .base_partial_3d_array import BasePartial3DArray
from .util import sanitise_slice, to_slice, unpack_slice, sanitise_unbounded_slice
from .data_types import DtypeType, Integer, IntegerType


class UnboundedPartial3DArray(BasePartial3DArray):
    """
    This is designed to work similarly to a numpy.ndarray but stores the data in a very different way.
    A numpy.ndarray stores a fixed size continuous array which for large arrays can become unmanageable.
    Sparse arrays allow individual values to exist which can be great where a small set of values are
    defined in a large area but get less efficient the denser the defined values are.

    This class was born out of the need for an array that has a fixed size in the horizontal directions but an
    unlimited height in the vertical distance (both above and below the origin)
    This is achieved by splitting the array into sections of fixed height and storing them sparsely so that only
    the defined sections need to be held in memory.

    This class implements an API that resembles that of a numpy array but it is not directly compatible with numpy.
    This class also implements methods to access and directly modify the underlying numpy arrays to give finer control.
    """

    def __init__(
        self,
        dtype: DtypeType,
        default_value: Union[int, bool],
        section_shape: Tuple[int, int, int],
        default_section_counts: Tuple[int, int],
        sections: Optional[Dict[int, numpy.ndarray]] = None,
    ):
        """
        Construct a :class:`UnboundedPartial3DArray`. This should not be used directly. You should use the relevant subclass for your use.

        :param dtype: The dtype that all arrays will be stored in.
        :param default_value: The default value that all undefined arrays will be populated with if required.
        :param section_shape: The shape of each section array.
        :param default_section_counts: A tuple containing the default number of sections above and below the origin in the y axis. This is used to define the default bounds.
        :param sections: The sections to initialise the array with.
        """
        super().__init__(
            dtype,
            default_value,
            section_shape,
            (0, None, 0),
            (section_shape[0], None, section_shape[0]),
            (1, 1, 1),
            sections=sections,
        )
        self._default_min_y = -default_section_counts[0] * self.section_shape[1]
        self._default_max_y = default_section_counts[1] * self.section_shape[1]

    def __repr__(self):
        return f"UnboundedPartial3DArray(dtype={self.dtype}, shape={self.shape})"

    @property
    def size_y(self) -> float:  # Literal[math.inf]:
        """The size of the array in the y axis. Is always :attr:`math.inf` for the unbounded variant. Read Only"""
        return math.inf

    @property
    def sections(self) -> Iterable[int]:
        """An iterable of the section defined in the array."""
        return self._sections.keys()

    def create_section(self, sy: IntegerType):
        """
        Create a section array at the given location using the default value and dtype.

        :param sy: The section index to create.
        """
        self._sections[int(sy)] = numpy.full(
            self.section_shape, self.default_value, dtype=self._dtype
        )

    def has_section(self, sy: int) -> bool:
        """Check if the array for a given section exists.
        :param cy: The section y index
        :return: True if the array exists, False otherwise
        """
        return sy in self._sections

    def add_section(self, sy: IntegerType, section: numpy.ndarray):
        """
        Add a section array at the given location.

        :param sy: The section index to assign the array to.
        :param section: The array to assign to the section. The shape must equal :attr:`section_shape`.
        """
        if section.shape != self._section_shape:
            raise ValueError(
                f"The size of all sections must be equal to the section_shape. Expected shape {self._section_shape}, got {section.shape}"
            )
        if section.dtype != self._dtype:
            section = section.astype(self._dtype)
        self._sections[int(sy)] = section

    def get_section(self, sy: Union[int, numpy.integer]) -> numpy.ndarray:
        """
        Get the section array for a given section index.

        If the section is not defined it will be populated using :meth:`create_section`

        :param sy: The section index to get.
        :return: Numpy array for this section
        """
        if sy not in self._sections:
            self.create_section(int(sy))
        return self._sections[sy]

    def __setitem__(
        self,
        slices: Tuple[
            Union[IntegerType, slice],
            Union[IntegerType, slice],
            Union[IntegerType, slice],
        ],
        value: Union[int, numpy.integer, numpy.ndarray],
    ):
        """
        Set a sub-section of the infinite height array.

        >>> # set the value at a given location
        >>> partial_array[3, 4, 5] = 1
        >>> # set a cuboid volume in the array
        >>> partial_array[2:3, 4:5, 6:7] = 1
        >>> # slice and int can be mixed
        >>> partial_array[2:3, 4, 6:7] = 1
        >>> # if an unbounded slice is given in the y axis it will be capped at the default max and min y values.
        >>> partial_array[2:3, :, 6:7] = 1

        :param slices: The slices or locations that define the volume to set.
        :param value: The value to set at the given location. Can be an integer or array.
        """
        if isinstance(value, Integer) and all(isinstance(s, Integer) for s in slices):
            sy, dy = self._section_index(slices[1])
            self.get_section(sy)[(slices[0], dy, slices[2])] = value
        else:
            self[slices][:, :, :] = value

    @overload
    def __getitem__(
        self, slices: Tuple[IntegerType, IntegerType, IntegerType]
    ) -> int: ...

    @overload
    def __getitem__(
        self,
        slices: Tuple[
            Union[IntegerType, slice],
            Union[IntegerType, slice],
            Union[IntegerType, slice],
        ],
    ) -> "BoundedPartial3DArray": ...

    def __getitem__(self, item):
        """
        Get a value or sub-section of the unbounded array.

        >>> # get the value at a given location
        >>> value = partial_array[3, 4, 5]  # an integer
        >>> # get a cuboid volume in the array
        >>> value = partial_array[2:3, 4:5, 6:7]  # BoundedPartial3DArray
        >>> # slice and int can be mixed
        >>> value = partial_array[2:3, 4, 6:7]  # BoundedPartial3DArray
        >>> # if an unbounded slice is given in the y axis it will be capped at the default max and min y values.
        >>> value = partial_array[2:3, :, 6:7]  # BoundedPartial3DArray

        :param item: The slices to extract.
        :return: The value or BoundedPartial3DArray viewing into this array.
        """
        if isinstance(item, tuple):
            if len(item) != 3:
                raise KeyError(f"Tuple item must be of length 3, got {len(item)}")
            if all(isinstance(i, Integer) for i in item):
                sy = item[1] // self.section_shape[1]
                if sy in self:
                    return int(
                        self._sections[sy][
                            (item[0], item[1] % self.section_shape[1], item[2])
                        ]
                    )
                else:
                    return self.default_value

            elif all(isinstance(i, (int, numpy.integer, slice)) for i in item):
                start_x, stop_x, step_x = sanitise_slice(
                    *unpack_slice(to_slice(item[0])), self.size_x
                )
                start_y, stop_y, step_y = sanitise_unbounded_slice(
                    *unpack_slice(to_slice(item[1])),
                    self._default_min_y,
                    self._default_max_y,
                )
                start_z, stop_z, step_z = sanitise_slice(
                    *unpack_slice(to_slice(item[2])), self.size_z
                )

                return BoundedPartial3DArray.from_partial_array(
                    self,
                    (start_x, start_y, start_z),
                    (stop_x, stop_y, stop_z),
                    (step_x, step_y, step_z),
                )
            else:
                raise KeyError(f"Unsupported tuple {item} for getitem")

        else:
            raise KeyError(
                f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
            )

    def __array__(self, dtype=None):
        """
        Get the data contained within as a numpy array.

        The y axis will be clamped to the default minimum and maximum y values.

        >>> numpy.array(partial_array)

        :param dtype: The dtype of the returned numpy array.
        :return: A numpy array of the contained data.
        """
        return self[:, :, :].__array__(dtype)


from .bounded_partial_3d_array import BoundedPartial3DArray



from __future__ import annotations
from typing import Dict, Iterable, List, Tuple, overload, Generator, Union
from numpy import integer

from amulet_nbt import ByteTag, IntTag, ShortTag, LongTag

from amulet.api.data_types import Int
from amulet.api.block import Block
from .base_registry import BaseRegistry


class BlockManager(BaseRegistry):
    """
    Class to handle the mappings between Block objects and their index-based internal IDs
    """

    def __init__(self, blocks: Iterable[Block] = ()):
        """
        Creates a new BlockManager object
        """
        self._index_to_block: List[Block] = []
        self._block_to_index_map: Dict[Block, int] = {}

        for block in blocks:
            assert isinstance(block, Block), "BlockManager only takes Block objects."
            if block not in self._block_to_index_map:
                self._block_to_index_map[block] = len(self._block_to_index_map)
            self._index_to_block.append(block)

    def __len__(self):
        """
        The number of blocks in the registry.

        >>> len(level.block_palette)
        10
        """
        return len(self._index_to_block)

    def __contains__(self, item: Block) -> bool:
        """
        Is the given :class:`Block` already in the registry.

        >>> block in level.block_palette
        True
        >>> 7 in level.block_palette
        True

        :param item: The block or index to check.
        """
        if isinstance(item, int):
            return item < len(self._index_to_block)
        elif isinstance(item, Block):
            return item in self._block_to_index_map
        return False

    def __iter__(self):
        """
        Iterate through all blocks in the registry.

        >>> for block in level.block_palette:
        >>>     ...
        """
        yield from self._index_to_block

    @property
    def blocks(self) -> Tuple[Block, ...]:
        """
        The blocks in the registry as a tuple.
        """
        return tuple(self._index_to_block)

    def values(self) -> Tuple[Block, ...]:
        """
        The blocks in the registry as a tuple.
        """
        return self.blocks

    def items(self) -> Generator[Tuple[int, Block], None, None]:
        """
        A generator of the block indexes and the block objects.
        """
        yield from enumerate(self._index_to_block)

    @overload
    def __getitem__(self, item: Block) -> int: ...

    @overload
    def __getitem__(self, item: Int) -> Block: ...

    @overload
    def __getitem__(
        self, item: Iterable[Union[Int, Block]]
    ) -> List[Union[Block, Int]]: ...

    def __getitem__(self, item):
        """
        If a Block object is passed to this function, it will return the internal ID/index of the blockstate.

        If an int is given, this method will return the Block object at that specified index.

        >>> level.block_palette[block]
        7
        >>> level.block_palette[7]
        block

        :param item: The Block object or int to get the mapping data of
        :return: An int if a Block object was supplied, a Block object if an int was supplied
        :raises KeyError if the requested item is not present.
        """
        try:
            return self._get_item(item)
        except (KeyError, IndexError):
            raise KeyError(
                f"There is no {item} in the BlockManager. "
                f"You might want to use the `get_add_block` function for your blocks before accessing them."
            )

    def _get_item(self, item):
        if isinstance(item, Block):
            return self._block_to_index_map[item]
        elif isinstance(item, (int, integer, ByteTag, ShortTag, IntTag, LongTag)):
            return self._index_to_block[int(item)]
        # if it isn't an Block or int assume an iterable of the above.
        return [self._get_item(i) for i in item]

    def get_add_block(self, block: Block) -> int:
        """
        Adds a Block object to the internal Block object/ID mappings.

        If the Block already exists in the mappings, the existing ID is returned.

        :param block: The Block to add to the manager
        :return: The internal ID of the Block
        """
        assert isinstance(block, Block), "BlockManager only takes Block objects."
        if block in self._block_to_index_map:
            return self._block_to_index_map[block]

        self._block_to_index_map[block] = i = len(self._block_to_index_map)
        self._index_to_block.append(block)

        return i

    def register(self, block: Block) -> int:
        """
        An alias of :meth:`get_add_block`.

        Adds a Block object to the internal Block object/ID mappings.

        If the Block already exists in the mappings, the existing ID is returned.

        :param block: The Block to add to the manager
        :return: The internal ID of the Block
        """
        return self.get_add_block(block)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk

from amulet_nbt import CompoundTag, ByteTag

from .base_anvil_interface import ChunkDataType, ChunkPathType

from .anvil_1912 import Anvil1912Interface as ParentInterface


class Anvil1934Interface(ParentInterface):
    """
    Made lighting optional
    """

    isLightOn: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("isLightOn", ByteTag)],
        ByteTag,
    )

    def __init__(self):
        super().__init__()
        self._set_feature("light_optional", "true")
        self._register_encoder(self._encode_is_light_on)
        self._register_decoder(self._decode_is_light_on)

    @staticmethod
    def minor_is_valid(key: int):
        return 1934 <= key < 2203

    def _decode_is_light_on(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["isLightOn"] = self.get_layer_obj(
            data, self.isLightOn, pop_last=True
        )

    def _encode_is_light_on(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        is_light_on = bool(chunk.misc.pop("isLightOn", None))
        self.set_layer_obj(data, self.isLightOn, ByteTag(is_light_on))


export = Anvil1934Interface



import os
import logging
import platformdirs

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


entity_support = False

IMG_DIRECTORY = os.path.join(os.path.dirname(__file__), "img")

# Initialise default paths. Applications should override these environment variables.
# os.environ.setdefault("DATA_DIR", platformdirs.user_data_dir("AmuletTeam", "AmuletTeam"))
# os.environ.setdefault("CONFIG_DIR", platformdirs.user_config_dir("AmuletTeam", "AmuletTeam"))
os.environ.setdefault(
    "CACHE_DIR", platformdirs.user_cache_dir("AmuletTeam", "AmuletTeam")
)
os.environ.setdefault("LOG_DIR", platformdirs.user_log_dir("AmuletTeam", "AmuletTeam"))

from .api import *
from amulet.level.load import load_level, load_format

# init a default logger
logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Tuple,
    Any,
    Union,
    TYPE_CHECKING,
    Optional,
    overload,
    Type,
    Sequence,
    Callable,
)
from enum import Enum

from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.data_types import AnyNDArray, VersionNumberAny, VersionIdentifierType
from amulet_nbt import (
    AbstractBaseTag,
    ListTag,
    CompoundTag,
    AnyNBT,
    NamedTag,
    StringTag,
    IntTag,
    IntArrayTag,
    DoubleTag,
    FloatTag,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from amulet.api.chunk import Chunk


class EntityIDType(Enum):
    int_id = 0
    str_id = 1
    namespace_str_id = 2
    namespace_str_Id = 3
    namespace_str_identifier = 4


class EntityCoordType(Enum):
    xyz_int = 0
    Pos_list_float = 1
    Pos_list_double = 2
    Pos_list_int = 3
    Pos_array_int = 4


PosTypeMap = {
    3: EntityCoordType.Pos_list_int,
    5: EntityCoordType.Pos_list_float,
    6: EntityCoordType.Pos_list_double,
}


class Interface(ABC):
    @abstractmethod
    def decode(self, *args, **kwargs) -> Tuple["Chunk", AnyNDArray]:
        raise NotImplementedError

    def _decode_entity(
        self,
        nbt: NamedTag,
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[Entity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return Entity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    def _decode_block_entity(
        self,
        nbt: NamedTag,
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[BlockEntity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return BlockEntity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    @staticmethod
    def _decode_base_entity(
        named_tag: NamedTag, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[
        Tuple[
            str,
            str,
            Union[int, float],
            Union[int, float],
            Union[int, float],
            NamedTag,
        ]
    ]:
        if not (
            isinstance(named_tag, NamedTag) and isinstance(named_tag.tag, CompoundTag)
        ):
            return

        nbt_tag = named_tag.tag

        if id_type == EntityIDType.namespace_str_id:
            entity_id = nbt_tag.pop("id", StringTag(""))
            if (
                not isinstance(entity_id, StringTag)
                or entity_id.py_str == ""
                or ":" not in entity_id.py_str
            ):
                return
            namespace, base_name = entity_id.py_str.split(":", 1)

        elif id_type == EntityIDType.namespace_str_Id:
            entity_id = nbt_tag.pop("Id", StringTag(""))
            if (
                not isinstance(entity_id, StringTag)
                or entity_id.py_str == ""
                or ":" not in entity_id.py_str
            ):
                return
            namespace, base_name = entity_id.py_str.split(":", 1)

        elif id_type == EntityIDType.str_id:
            entity_id = nbt_tag.pop("id", StringTag(""))
            if not isinstance(entity_id, StringTag) or entity_id.py_str == "":
                return
            namespace = ""
            base_name = entity_id.py_str

        elif id_type in [EntityIDType.namespace_str_identifier, EntityIDType.int_id]:
            if "identifier" in nbt_tag:
                entity_id = nbt_tag.pop("identifier")
                if (
                    not isinstance(entity_id, StringTag)
                    or entity_id.py_str == ""
                    or ":" not in entity_id.py_str
                ):
                    return
                namespace, base_name = entity_id.py_str.split(":", 1)
            elif "id" in nbt_tag:
                entity_id = nbt_tag.pop("id")
                if not isinstance(entity_id, IntTag):
                    return
                namespace = ""
                base_name = str(entity_id.py_int)
            else:
                return
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type in [
            EntityCoordType.Pos_list_double,
            EntityCoordType.Pos_list_float,
            EntityCoordType.Pos_list_int,
        ]:
            if "Pos" not in nbt_tag:
                return
            pos = nbt_tag.pop("Pos")
            pos: ListTag

            if (
                not isinstance(pos, ListTag)
                or len(pos) != 3
                or PosTypeMap.get(pos.list_data_type) != coord_type
            ):
                return
            x, y, z = [c.py_data for c in pos]
        elif coord_type == EntityCoordType.Pos_array_int:
            if "Pos" not in nbt_tag:
                return
            pos = nbt_tag.pop("Pos")
            pos: IntArrayTag

            if not isinstance(pos, IntArrayTag) or len(pos) != 3:
                return
            x, y, z = pos
        elif coord_type == EntityCoordType.xyz_int:
            if not all(
                c in nbt_tag and isinstance(nbt_tag[c], IntTag) for c in ("x", "y", "z")
            ):
                return
            x, y, z = [nbt_tag.pop(c).py_int for c in ("x", "y", "z")]
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return namespace, base_name, x, y, z, named_tag

    @abstractmethod
    def encode(self, *args, **kwargs) -> Any:
        """
        Take a version-specific chunk and encode it to raw data for the format to store.
        """
        raise NotImplementedError

    def _encode_entity(
        self, entity: Entity, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[NamedTag]:
        return self._encode_base_entity(entity, id_type, coord_type)

    def _encode_block_entity(
        self, entity: BlockEntity, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[NamedTag]:
        return self._encode_base_entity(entity, id_type, coord_type)

    @staticmethod
    def _encode_base_entity(
        entity: Union[Entity, BlockEntity],
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[NamedTag]:
        if not isinstance(entity.nbt, NamedTag) and isinstance(
            entity.nbt.tag, CompoundTag
        ):
            return
        named_tag = entity.nbt
        tag = named_tag.compound

        if id_type == EntityIDType.namespace_str_id:
            tag["id"] = StringTag(entity.namespaced_name)
        elif id_type == EntityIDType.namespace_str_Id:
            tag["Id"] = StringTag(entity.namespaced_name)
        elif id_type == EntityIDType.namespace_str_identifier:
            tag["identifier"] = StringTag(entity.namespaced_name)
        elif id_type == EntityIDType.str_id:
            tag["id"] = StringTag(entity.base_name)
        elif id_type == EntityIDType.int_id:
            if not entity.base_name.isnumeric():
                return
            tag["id"] = IntTag(int(entity.base_name))
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type == EntityCoordType.Pos_list_double:
            tag["Pos"] = ListTag(
                [
                    DoubleTag(float(entity.x)),
                    DoubleTag(float(entity.y)),
                    DoubleTag(float(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_list_float:
            tag["Pos"] = ListTag(
                [
                    FloatTag(float(entity.x)),
                    FloatTag(float(entity.y)),
                    FloatTag(float(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_list_int:
            tag["Pos"] = ListTag(
                [
                    IntTag(int(entity.x)),
                    IntTag(int(entity.y)),
                    IntTag(int(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_array_int:
            tag["Pos"] = IntArrayTag([int(entity.x), int(entity.y), int(entity.z)])
        elif coord_type == EntityCoordType.xyz_int:
            tag["x"] = IntTag(int(entity.x))
            tag["y"] = IntTag(int(entity.y))
            tag["z"] = IntTag(int(entity.z))
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return named_tag

    @overload
    @staticmethod
    def check_type(obj: CompoundTag, key: str, dtype: Type[AnyNBT]) -> bool: ...

    @overload
    @staticmethod
    def check_type(obj: ListTag, key: int, dtype: Type[AnyNBT]) -> bool: ...

    @staticmethod
    def check_type(
        obj: Union[CompoundTag, ListTag], key: Union[str, int], dtype: Type[AnyNBT]
    ) -> bool:
        """Check the key exists and the type is correct."""
        return key in obj and isinstance(obj[key], dtype)

    @overload
    def get_obj(
        self,
        obj: CompoundTag,
        key: str,
        dtype: Type[AnyNBT],
        default: Optional[AnyNBT] = None,
    ) -> Optional[AnyNBT]: ...

    @overload
    def get_obj(
        self,
        obj: ListTag,
        key: int,
        dtype: Type[AnyNBT],
        default: Optional[AnyNBT] = None,
    ) -> Optional[AnyNBT]: ...

    def get_obj(
        self, obj, key, dtype: Type[AnyNBT], default: Optional[AnyNBT] = None
    ) -> Optional[AnyNBT]:
        """Pop a key from a container object if it exists and the type is correct. Otherwise return default.
        This works in much the same way as dict.get but uses default if the data type does not match.

        :param obj: The CompoundTag to read from.
        :param key: The key to use.
        :param dtype: The expected data type.
        :param default: The default value to use if the existing is not valid. If None will use dtype()
        :return: The final value in the key.
        """
        if key in obj:
            if isinstance(obj[key], dtype):
                # if it exists and is correct
                return obj.pop(key)
        if default is None:
            return dtype()
        return default

    def get_nested_obj(
        self,
        obj: Union[CompoundTag, ListTag],
        path: Sequence[  # The path to the object.
            Tuple[Union[str, int], Type[AbstractBaseTag]],
        ],
        default: Union[None, AnyNBT, Callable[[], Any]] = None,
        *,
        pop_last=False,
    ):
        """
        Get an object from a nested NBT structure

        :param obj: The root NBT object
        :param path: The path to the desired object (key, dtype)
        :param default: The default value to use if the existing is not valid. If default is callable then return the called result.
        :param pop_last: If true the last key will be popped
        :return:
        """
        try:
            last_index = len(path) - 1
            for i, (key, dtype) in enumerate(path):
                if i == last_index and pop_last:
                    obj = obj.pop(key)
                else:
                    obj = obj[key]
                if dtype is not obj.__class__:
                    raise TypeError
        except (KeyError, IndexError, TypeError):
            if default is None or isinstance(default, AbstractBaseTag):
                return default
            elif callable(default):
                return default()
            else:
                raise TypeError(
                    "default must be None, an NBT instance or an NBT class."
                )
        else:
            return obj

    @staticmethod
    def set_obj(
        obj: CompoundTag,
        key: str,
        dtype: Type[AnyNBT],
        default: Union[None, AnyNBT, Callable[[], Any]] = None,
        path: Sequence[str] = (),  # The path to the object.
        *,
        setdefault=False,
    ) -> AnyNBT:
        """
        Works like setdefualt on a dictionary but works with an optional nested path.

        :param obj: The compound tag to get the data from
        :param key: The key to setdefault
        :param dtype: The dtype that the key must be
        :param default: The default value to set if it does not exist or the type is wrong
        :param path: Optional path to the nested compound.
        :param setdefault: If True will behave like setdefault. If False will replace existing data.
        :return: The data at the path
        """
        for path_key in path:
            obj_ = obj.get(path_key, None)
            if not isinstance(obj_, CompoundTag):
                # if it does not exist or the type is wrong then create it
                obj_ = obj[path_key] = CompoundTag()
            obj = obj_
        obj_ = obj.get(key, None)
        if not setdefault or not isinstance(obj_, dtype):
            # if it does not exist or the type is wrong then create it
            if default is None:
                obj_ = dtype()
            elif isinstance(default, AbstractBaseTag):
                obj_ = default
            elif callable(default):
                obj_ = default()
            else:
                raise TypeError(
                    "default must be None, an NBT instance or an NBT class."
                )
            obj[key] = obj_
        return obj_

    @abstractmethod
    def get_translator(
        self,
        max_world_version: VersionIdentifierType,
        data: Any = None,
    ) -> Tuple["Translator", VersionNumberAny]:
        """
        Get the Translator class for the requested version.
        :param max_world_version: The game version the world was last opened in. Version number tuple or data version number.
        :param data: Optional data to get translator based on chunk version rather than world version
        :return: Tuple[Translator, version number for PyMCTranslate to use]
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def is_valid(key: Tuple) -> bool:
        """
        Returns whether this Interface is able to interface with the chunk type with a given identifier key,
        generated by the format.

        :param key: The key who's decodability needs to be checked.
        :return: True if this interface can interface with the chunk version associated with the key, False otherwise.
        """
        raise NotImplementedError



from __future__ import annotations

from amulet.api.chunk import Chunk

from .base_anvil_interface import ChunkDataType
from .anvil_1503 import Anvil1503Interface as ParentInterface


class Anvil1519Interface(ParentInterface):
    """
    Sub-chunk sections must have a block array if defined
    """

    def __init__(self):
        super().__init__()
        self._register_post_encoder(self._post_encode_sections)

    @staticmethod
    def minor_is_valid(key: int):
        return 1519 <= key < 1901

    def _post_encode_sections(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        sections = self.get_layer_obj(data, self.Sections)
        for section_index in range(len(sections) - 1, -1, -1):
            if (
                "BlockStates" not in sections[section_index]
                or "Palette" not in sections[section_index]
            ):
                del sections[section_index]


export = Anvil1519Interface



from collections import UserList
from typing import Iterable
from amulet.api.entity import Entity


class ChunkList(UserList):
    def __init__(self, iterable: Iterable = ()):
        super().__init__(list(iterable))

    def _check_all_types(self, value):
        for val in value:
            self._check_type(val)

    def _check_type(self, value):
        raise NotImplementedError

    def append(self, value) -> None:
        """Append value to the end of the list."""
        self._check_type(value)
        super().append(value)

    def copy(self) -> "ChunkList":
        raise NotImplementedError

    def extend(self, iterable: Iterable):
        """Extend list by appending elements from the iterable."""
        self._check_all_types(iterable)
        super().extend(iterable)

    def insert(self, index: int, value):
        """Insert value before index."""
        self._check_type(value)
        super().insert(index, value)

    def __repr__(self):
        if not self:
            return "[]"
        return "[" + "".join(f"\n\t{str(obj)}" for obj in self) + "\n]"


class EntityList(ChunkList):
    """
    A custom implementation of the list class.

    It can only store :class:`Entity` instances.
    """

    def __init__(self, entities: Iterable[Entity] = ()):
        """
        Construct an :class:`EntityList`.

        :param entities: An Iterable of :class:`Entity` objects.
        """
        for e in entities:
            self._check_type(e)
        super().__init__(entities)

    def _check_type(self, value):
        assert isinstance(value, Entity), f"{value} is not an instance of Entity"

    def copy(self) -> "EntityList":
        """
        Create a shallow copy of the entity container.
        """
        return EntityList(self)

    def __iter__(self) -> Iterable[Entity]:
        """
        An iterable of all the :class:`Entity` objects.
        """
        yield from self.data

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"EntityList({super().__repr__()})"



from amulet_nbt import CompoundTag, StringTag


class Item(CompoundTag):
    """
    {
        "namespace": StringTag,
        "base_name": StringTag,
        "metadata": CompoundTag
    }
    """

    def __init__(self, namespace: str, base_name: str, metadata: dict = None):
        super().__init__(
            {
                "namespace": StringTag(namespace),
                "base_name": StringTag(base_name),
                "metadata": CompoundTag(metadata or {}),
            }
        )

    @property
    def namespace(self) -> str:
        return self.get_string("namespace").py_str

    @property
    def base_name(self) -> str:
        return self.get_string("base_name").py_str

    @property
    def metadata(self) -> CompoundTag:
        return self.get_compound("metadata")


class BlockItem(CompoundTag):
    """
    {
        "namespace": StringTag,
        "base_name": StringTag,
        "properties": CompoundTag
        "metadata": CompoundTag
    }
    """

    def __init__(
        self,
        namespace: str,
        base_name: str,
        properties: dict = None,
        metadata: dict = None,
    ):
        super().__init__(
            {
                "namespace": StringTag(namespace),
                "base_name": StringTag(base_name),
                "properties": CompoundTag(properties or {}),
                "metadata": CompoundTag(metadata or {}),
            }
        )

    @property
    def namespace(self) -> str:
        return self.get_string("namespace").py_str

    @property
    def base_name(self) -> str:
        return self.get_string("base_name").py_str

    @property
    def properties(self) -> CompoundTag:
        return self.get_compound("properties")

    @property
    def metadata(self) -> CompoundTag:
        return self.get_compound("metadata")
